from __future__ import division

from inferelator import utils
from inferelator import default
from inferelator.preprocessing.metadata_parser import MetadataHandler
from inferelator.preprocessing.metadata_parser import ConditionDoesNotExistError, MultipleConditionsError
import pandas as pd
import numpy as np


class PythonDRDriver(object):
    # Parameters for response matrix
    tau = default.DEFAULT_TAU
    delTmin = default.DEFAULT_DELTMIN
    delTmax = default.DEFAULT_DELTMAX

    # Strict checking will raise exception if there are potential inconsistencies
    # Strict_checking_for_metadata raises an exception if there are any missing experiments in the metadata
    # Strict_checking_for_duplicates raises an exception if there are non-trivial duplicates in the metadata
    strict_checking_for_metadata = False
    strict_checking_for_duplicates = True

    # Deep walking timecourse experiments means generating separate conditions for every linked experiment within the
    # acceptable time window. If set to false, it will just use the most recent linked experiment.
    deep_walk_timecourse_exps = False

    # Set return_half_tau to true to return the half_tau_response_matrix
    return_half_tau = False

    # Expression data
    sample_names = None

    # Metadata
    metadata_handler = "branching"
    steady_idx = None
    ts_group = None

    def __init__(self, metadata_handler=None, tau=None, deltmin=None, deltmax=None, return_half_tau=False):
        self.tau = tau if tau is not None else default.DEFAULT_TAU
        self.delTmin = deltmin if deltmin is not None else default.DEFAULT_DELTMIN
        self.delTmax = deltmax if deltmax is not None else default.DEFAULT_DELTMAX
        self.return_half_tau = return_half_tau
        self.metadata_handler = metadata_handler if metadata_handler is not None else self.metadata_handler

    def validate_run(self, meta_data):
        """
        Return False if the metadata provided will not work with this module. Return True otherwise
        :param meta_data: pd.DataFrame [N x 5]
        :return: bool
        """
        return MetadataHandler.get_handler(self.metadata_handler).validate_metadata_columns(meta_data)

    def run(self, exp_data, meta_data):
        """
        Process expression data and metadata into design & response data
        :param exp_data: pd.DataFrame [G x N]
        :param meta_data: pd.DataFrame [N x 5]
        :return design, response: pd.DataFrame [G x N], pd.DataFrame [G x N]
        """

        (k, n) = exp_data.shape
        processor = MetadataHandler.get_handler(self.metadata_handler)

        # Turn NA in the dataframe into np.NaN
        meta_data = processor.fix_NAs(meta_data)

        # Validate metadata alignment to expression
        processor.validate_metadata(exp_data, meta_data)

        # Turn the metadata into a set of dicts keyed by sample
        steady_idx, ts_group = processor.process_groups(meta_data)

        # Check and make sure that the metadata matches experimental data and whatnot
        steady_idx = processor.check_for_dupes(exp_data, meta_data, steady_idx,
                                               strict_checking_for_metadata=self.strict_checking_for_metadata,
                                               strict_checking_for_duplicates=self.strict_checking_for_duplicates)

        # Pull apart the expression dataframe into indexes and an ndarray
        genes = exp_data.index.values
        self.sample_names = exp_data.columns.values.astype(str)
        exp_data = exp_data.values.astype(np.dtype('float64'))

        # Construct empty arrays for the output data
        col_labels = []
        included = np.zeros((n, 1), dtype=bool)
        design = []
        response = []
        response_half = []

        # Walk through all the conditions in the expression data
        for c_idx, cc in enumerate(self.sample_names):
            utils.Debug.vprint("Processing condition {cc} [{c} / {tot}]".format(cc=cc, c=c_idx + 1, tot=n), level=3)
            if steady_idx[cc]:
                # This is a steady-state experiment
                self.static_exp(c_idx, cc, col_labels, included, exp_data, design, response, response_half)
            else:
                # This is a timecourse experiment
                for prev_cond, prev_delt in self._get_prior_timepoints(ts_group, cc):
                    prev_idx = self._get_index(self.sample_names, prev_cond)
                    self.timecourse_exp(c_idx, prev_idx, prev_delt, col_labels, included, exp_data, design, response,
                                        response_half)
                    if not self.deep_walk_timecourse_exps:
                        break

        for c_idx in np.where(~included)[0].tolist():
            # Run anything that wasn't included initially in as a steady-state experiment
            cc = self.sample_names[c_idx]
            self.static_exp(c_idx, cc, col_labels, included, exp_data, design, response, response_half)

        design = pd.DataFrame(np.array(design), index=col_labels, columns=genes).transpose()
        response = pd.DataFrame(np.array(response), index=col_labels, columns=genes).transpose()

        if self.return_half_tau:
            response_half = pd.DataFrame(np.array(response_half), index=col_labels, columns=genes).transpose()
            return design, response, response_half
        else:
            return design, response

    @staticmethod
    def static_exp(c_idx, c_name, col_labels, included, exp_data, design, response, response_half):
        """
        Concatenate expression data onto design, response & response half

        :param c_idx: int
        :param c_name: str
        :param col_labels: list
        :param included: dict
        :param exp_data: np.ndarray
        :param design: list(np.ndarray)
        :param response: list(np.ndarray)
        :param response_half: list(np.ndarray)
        """
        col_labels.append(c_name)
        included[c_idx] = True
        design.append(exp_data[:, c_idx].flatten())
        response.append(exp_data[:, c_idx].flatten())
        response_half.append(exp_data[:, c_idx].flatten())

    def timecourse_exp(self, c_idx, prev_idx, prev_delt, col_labels, included, exp_data, design, response,
                       response_half):
        """
        Concatenate expression data from the prior timepoint onto design.
        Calculate response data based on timecourse and concatenate the result onto response & response half.
        :param c_idx: int
        :param prev_idx: int
        :param prev_delt: numeric
        :param col_labels: list
        :param included: dict
        :param exp_data: np.ndarray
        :param design: list(np.ndarray)
        :param response: list(np.ndarray)
        :param response_half: list(np.ndarray)
        """

        col_labels.append(str(self.sample_names[prev_idx]) + "-" + str(self.sample_names[c_idx]))
        included[c_idx] = True
        included[prev_idx] = True

        resp, half_resp = self._calculate_ts_response(exp_data, self.tau, prev_delt, c_idx, prev_idx)

        design.append(exp_data[:, prev_idx].flatten())
        response.append(resp.flatten())
        response_half.append(half_resp.flatten())

    def _get_prior_timepoints(self, ts_group, cond):
        """
        Walk backwards through timepoints until a total del.t that falls within the acceptable window is located
        :param cond:
        :return:
        """
        total_delt = 0
        for pcond, pdelt in self._prior_timepoint_generator(ts_group, cond):
            total_delt += pdelt
            if self.delTmax is not None:
                if self.delTmin <= total_delt <= self.delTmax:
                    yield pcond, total_delt
                elif total_delt > self.delTmax:
                    break
            else:
                if self.delTmin <= total_delt:
                    yield pcond, total_delt

    @staticmethod
    def _calculate_ts_response(exp_data, tau, prev_delt, idx, prev_idx):
        diff = exp_data[:, idx] - exp_data[:, prev_idx]
        resp = float(tau) / float(prev_delt) * diff + exp_data[:, prev_idx]
        half_resp = float(tau) / 2 / float(prev_delt) * diff + exp_data[:, prev_idx]
        return resp, half_resp

    @staticmethod
    def _prior_timepoint_generator(ts_group, cond):
        """
        Yield the previous timepoint condition name and delt until it gets back to the first timepoint
        :param cond:
        """
        prev_cond, prev_delt = ts_group[cond][0]
        while prev_cond is not None:
            yield prev_cond, prev_delt
            prev_cond, prev_delt = ts_group[prev_cond][0]

    @staticmethod
    def _get_index(sample_names, cond):
        """
        Look up the index in the expression data of a specific condition. Raise errors if it doesn't exist, or if it's
        not unique
        :param sample_names: list
        :param cond: str
        :return idx: int
        """
        idx = np.where(sample_names == cond)[0].tolist()
        if len(idx) == 0:
            raise ConditionDoesNotExistError("{cond} cannot be identified in expression conditions".format(cond=cond))
        if len(idx) > 1:
            raise MultipleConditionsError("{cond} is not unique in expression conditions".format(cond=cond))
        else:
            idx = idx[0]
        return idx



from inferelator.workflow import _H5AD, _HDF5, _TSV
from inferelator.tfa_workflow import TFAWorkFlow
from inferelator.single_cell_workflow import SingleCellWorkflow
from inferelator.utils import InferelatorDataLoader, InferelatorData, Debug, Validator as check
from inferelator.preprocessing.velocity_tfa import VelocityTFA
import numpy as np


_VELOCITY_FILE_TYPES = [_TSV, _HDF5, _H5AD]


class VelocityWorkflow(SingleCellWorkflow):
    _velocity_data = None
    _velocity_file_name = None
    _velocity_file_type = None
    _velocity_h5_layer = None

    _decay_constants = None
    _use_precalculated_decay_constants = True

    tau = None
    tfa_driver = VelocityTFA

    def get_data(self):
        super(VelocityWorkflow, self).get_data()
        self.load_velocity()

    def startup_finish(self):
        self.single_cell_normalize()
        self._align_velocity()
        TFAWorkFlow.startup_finish(self)

    def set_velocity_parameters(self, velocity_file_name=None, velocity_file_type=None, velocity_file_layer=None):
        """
        Set velocity file arguments

        :param velocity_file_name: File name that has velocity data. Orientation must match expression data
        :type velocity_file_name: str
        :param velocity_file_type: Type of file to load. Accepts 'tsv', 'h5ad' and 'hdf5'.
        :type velocity_file_type: str
        :param velocity_file_layer: If the loaded file is an h5 file, which layer should be used
        :type velocity_file_layer: str
        """

        self._set_with_warning("_velocity_file_name", velocity_file_name)
        self._set_with_warning("_velocity_h5_layer", velocity_file_layer)

        if velocity_file_type is not None and velocity_file_type.lower() in _VELOCITY_FILE_TYPES:
            self._set_with_warning("_velocity_file_type", velocity_file_type)
        elif velocity_file_type is not None:
            msg = "velocity_file_type must be in {ft}".format(ft=_VELOCITY_FILE_TYPES)
            raise ValueError(msg)

    def load_velocity(self, velocity_file=None, loader_type=None):

        velocity_file = self._velocity_file_name if velocity_file is None else velocity_file
        loader_type = self._velocity_file_type if loader_type is None else loader_type
        transpose = not self.expression_matrix_columns_are_genes

        loader = InferelatorDataLoader(input_dir=self.input_dir, file_format_settings=self._file_format_settings)
        Debug.vprint("Loading velocity data from {f}".format(f=velocity_file), level=1)

        if loader_type == _TSV or loader_type is None:
            self._velocity_data = loader.load_data_tsv(velocity_file, transpose_expression_data=transpose)

        elif loader_type == _H5AD:
            self._velocity_data = loader.load_data_h5ad(velocity_file, use_layer=self._velocity_h5_layer)

        elif loader_type == _HDF5:
            self._velocity_data = loader.load_data_hdf5(velocity_file, transpose_expression_data=transpose,
                                                        use_layer=self._velocity_h5_layer)
        else:
            raise ValueError("Invalid velocity_file_type: {a}".format(a=loader_type))

        self._velocity_data.name = "Velocity"

    def _align_velocity(self):

        keep_genes = self._velocity_data.gene_names.intersection(self.data.gene_names)
        Debug.vprint("Aligning velocity and expression data on {n} genes".format(n=len(keep_genes)))

        self._velocity_data.trim_genes(remove_constant_genes=False, trim_gene_list=keep_genes)
        self.data.trim_genes(remove_constant_genes=False, trim_gene_list=keep_genes)

        assert check.indexes_align((self._velocity_data.gene_names, self.data.gene_names))
        assert check.indexes_align((self._velocity_data.sample_names, self.data.sample_names))

    def compute_common_data(self):
        pass

    def _recalculate_design(self):
        """
        Calculate dX/dt + lambda * X as response and A_hat as design
        :return:
        """

        self.response = self._combine_expression_velocity(self.data, self._velocity_data)
        self.data = None
        self._velocity_data = None

        self.design = self.tfa_driver().compute_transcription_factor_activity(self.priors_data, self.response)

    def _combine_expression_velocity(self, expression, velocity):
        """
        Calculate dX/dt + lambda * X
        :param expression:
        :param velocity:
        :return:
        """

        assert check.indexes_align((expression.gene_names, velocity.gene_names))
        assert check.indexes_align((expression.sample_names, velocity.sample_names))

        if self._decay_constants is not None:
            Debug.vprint("Using preloaded decay constants in _decay_constants")
            decay_constants = self._decay_constants
        elif self.tau is not None:
            Debug.vprint("Calculating decay constants for tau {t}".format(t=self.tau))
            decay_constants = np.repeat(1 / self.tau, expression.num_genes)
        elif "decay_constants" in velocity.gene_data.columns and self._use_precalculated_decay_constants:
            Debug.vprint("Extracting decay constants from {n}".format(n=velocity.name))
            decay_constants = velocity.gene_data["decay_constants"].values
        elif "decay_constants" in expression.gene_data.columns and self._use_precalculated_decay_constants:
            Debug.vprint("Extracting decay constants from {n}".format(n=expression.name))
            decay_constants = expression.gene_data["decay_constants"].values
        else:
            Debug.vprint("No decay information found. Solving dX/dt = AB for Betas")
            return velocity

        x = np.multiply(expression.values, decay_constants[None, :])
        return InferelatorData(np.add(velocity.values, x), gene_names=expression.gene_names,
                               sample_names=expression.sample_names, meta_data=expression.meta_data)



from inferelator.regression import sklearn_regression
from sklearn.linear_model import ElasticNetCV
import copy

ELASTICNET_PARAMETERS = dict(l1_ratio=[0.5, 0.7, 0.9],
                             eps=0.001,
                             n_alphas=50,
                             alphas=None,
                             fit_intercept=True,
                             normalize=False,
                             precompute='auto',
                             max_iter=1000,
                             tol=0.001,
                             cv=3,
                             copy_X=True,
                             verbose=0,
                             n_jobs=1,
                             positive=False,
                             selection='random',
                             min_coef=0.1)


class ElasticNetWorkflowMixin(sklearn_regression.SKLearnWorkflowMixin):
    """
    Set default parameters to run scikit-learn ElasticNetCV
    """

    _sklearn_model = ElasticNetCV
    _sklearn_add_random_state = True

    def __init__(self, *args, **kwargs):
        super(ElasticNetWorkflowMixin, self).__init__(*args, **kwargs)
        self._sklearn_model_params = copy.copy(ELASTICNET_PARAMETERS)


class ElasticNetByTaskRegressionWorkflowMixin(sklearn_regression.SKLearnByTaskMixin):

    _sklearn_model = ElasticNetCV
    _sklearn_add_random_state = True

    def __init__(self, *args, **kwargs):
        super(ElasticNetByTaskRegressionWorkflowMixin, self).__init__(*args, **kwargs)
        self._sklearn_model_params = copy.copy(ELASTICNET_PARAMETERS)



from __future__ import print_function, unicode_literals, division
import os

from inferelator.default import SBATCH_VARS


class Debug:
    """
    This class is for printing status messages to stdout
    Just plain print doesn't work so well when there are multiple processes
    """
    verbose_level = 0
    default_level = 1

    silence_clients = True
    is_master = True

    levels = dict(silent=-1,
                  normal=0,
                  verbose=1, v=1,
                  very_verbose=2, vv=2,
                  max_output=3, vvv=3)

    @classmethod
    def set_verbose_level(cls, lvl):
        if isinstance(lvl, (int, float)):
            cls.verbose_level = lvl
        elif lvl in cls.levels.keys():
            cls.verbose_level = cls.levels[lvl]

    @classmethod
    def vprint(cls, *args, **kwargs):
        if cls.silence_clients and not cls.is_master:
            return
        cls.print_level(*args, **kwargs)

    @classmethod
    def allprint(cls, *args, **kwargs):
        cls.print_level(*args, **kwargs)

    @classmethod
    def print_level(cls, *args, **kwargs):
        try:
            level = kwargs.pop('level')
        except KeyError:
            level = cls.default_level
        if level <= cls.verbose_level:
            print((" " * level), *args, **kwargs)
        else:
            return


def inferelator_verbose_level(level):
    """
    Set verbosity.
    :param level: Verbose level. 0 is normal, 1 is extra information. 2+ is not recommended. -1 silences most outputs.
    :type level: int
    """
    Debug.set_verbose_level(level)


def slurm_envs(var_names=None):
    """
    Get environment variable names and return them as a dict
    :param var_names: list
        A list of environment variable names to get. Will throw an error if they're not keys in the SBATCH_VARS dict
    :return envs: dict
        A dict keyed by setattr variable name of the value (or default) from the environment variables
    """
    var_names = SBATCH_VARS.keys() if var_names is None else var_names
    assert set(var_names).issubset(set(SBATCH_VARS.keys()))

    envs = {}
    for cv in var_names:
        os_var, mt, de = SBATCH_VARS[cv]
        try:
            val = mt(os.environ[os_var])
        except (KeyError, TypeError):
            val = de
        envs[cv] = val
    return envs


from inferelator.utils.validator import Validator, is_string
from inferelator.utils.debug import Debug, slurm_envs, inferelator_verbose_level
from inferelator.utils.loader import InferelatorDataLoader, DEFAULT_PANDAS_TSV_SETTINGS
from inferelator.utils.data import (InferelatorData, df_from_tsv, array_set_diag, df_set_diag,
                                    melt_and_reindex_dataframe, make_array_2d, scale_vector, DotProduct )




import numpy as np
from inferelator.utils import Validator as check
from inferelator import utils
from inferelator.regression import base_regression
from inferelator.distributed.inferelator_mp import MPControl
from sklearn.base import BaseEstimator
from inferelator.regression.base_regression import _MultitaskRegressionWorkflowMixin
import copy
import inspect


def sklearn_gene(x, y, model, min_coef=None, **kwargs):
    """
    Use a scikit-learn model for regression

    :param x: Feature array
    :type x: np.ndarray [N x K]
    :param y: Response array
    :type y: np.ndarray [N x 1]
    :param model: Instance of a scikit BaseEstimator-derived model
    :type model: BaseEstimator
    :param min_coef: A minimum coefficient value to include in the model. Any values smaller will be set to 0.
    :type min_coef: numeric
    :return: A dict of results for this gene
    :rtype: dict
    """
    assert check.argument_type(x, np.ndarray)
    assert check.argument_type(y, np.ndarray)
    assert check.argument_is_subclass(model, BaseEstimator)

    (N, K) = x.shape

    # Fit the model
    model.fit(x, y, **kwargs)

    # Get all model coefficients [K, ]
    try:
        coefs = model.coef_
    except AttributeError:
        coefs = model.estimator_.coef_

    # Set coefficients below threshold to 0
    if min_coef is not None:
        coefs[np.abs(coefs) < min_coef] = 0.  # Threshold coefficients

    coef_nonzero = coefs != 0  # Create a boolean array where coefficients are nonzero [K, ]

    # If there are non-zero coefficients, redo the linear regression with them alone
    # And calculate beta_resc
    if coef_nonzero.sum() > 0:
        x = x[:, coef_nonzero]
        utils.make_array_2d(y)
        betas = base_regression.recalculate_betas_from_selected(x, y)
        betas_resc = base_regression.predict_error_reduction(x, y, betas)
        return dict(pp=coef_nonzero,
                    betas=betas,
                    betas_resc=betas_resc)
    else:
        return dict(pp=np.repeat(True, K).tolist(),
                    betas=np.zeros(K),
                    betas_resc=np.zeros(K))


class SKLearnRegression(base_regression.BaseRegression):

    def __init__(self, x, y, model, random_state=None, **kwargs):
        self.params = kwargs

        if random_state is not None:
            self.params["random_state"] = random_state

        self.min_coef = self.params.pop("min_coef", None)
        self.model = model(**self.params)

        super(SKLearnRegression, self).__init__(x, y)

    def regress(self):
        """
        Execute Elastic Net

        :return: list
            Returns a list of regression results that base_regression's pileup_data can process
        """

        if MPControl.is_dask():
            from inferelator.distributed.dask_functions import sklearn_regress_dask
            return sklearn_regress_dask(self.X, self.Y, self.model, self.G, self.genes, self.min_coef)

        def regression_maker(j):
            level = 0 if j % 100 == 0 else 2
            utils.Debug.allprint(base_regression.PROGRESS_STR.format(gn=self.genes[j], i=j, total=self.G), level=level)

            data = sklearn_gene(self.X.values,
                                utils.scale_vector(self.Y.get_gene_data(j, force_dense=True).flatten()),
                                copy.copy(self.model),
                                min_coef=self.min_coef)
            data['ind'] = j
            return data

        return MPControl.map(regression_maker, range(self.G), tell_children=False)


class SKLearnWorkflowMixin(base_regression._RegressionWorkflowMixin):
    """
    Use any scikit-learn regression module
    """

    _sklearn_model = None
    _sklearn_model_params = None
    _sklearn_add_random_state = False

    def __init__(self, *args, **kwargs):
        self._sklearn_model_params = {}
        super(SKLearnWorkflowMixin, self).__init__(*args, **kwargs)

    def set_regression_parameters(self, model=None, add_random_state=None, **kwargs):
        """
        Set parameters to use a sklearn model for regression

        :param model: A scikit-learn model class
        :type model: BaseEstimator subclass
        :param add_random_state: Flag to include workflow random seed as "random_state" in the model
        :type add_random_state: bool
        :param kwargs: Any arguments which should be passed to the scikit-learn model class instantiation
        :type kwargs: any
        """

        if model is not None and not inspect.isclass(model):
            raise ValueError("Pass an uninstantiated scikit-learn model (i.e. LinearRegression, not LinearRegression()")

        self._set_with_warning("_sklearn_model", model)
        self._set_without_warning("_sklearn_add_random_state", add_random_state)
        self._sklearn_model_params.update(kwargs)

    def run_bootstrap(self, bootstrap):
        x = self.design.get_bootstrap(bootstrap)
        y = self.response.get_bootstrap(bootstrap)
        utils.Debug.vprint('Calculating betas using SKLearn model {m}'.format(m=self._sklearn_model.__name__), level=0)
        MPControl.sync_processes(pref="sk_pre")
        return SKLearnRegression(x,
                                 y,
                                 self._sklearn_model,
                                 random_state=self.random_seed if self._sklearn_add_random_state else None,
                                 **self._sklearn_model_params).run()


class SKLearnByTaskMixin(_MultitaskRegressionWorkflowMixin, SKLearnWorkflowMixin):
    """
    This runs BBSR regression on tasks defined by the AMUSR regression (MTL) workflow
    """

    def run_bootstrap(self, bootstrap_idx):
        betas, betas_resc = [], []

        # Select the appropriate bootstrap from each task and stash the data into X and Y
        for k in range(self._n_tasks):
            x = self._task_design[k].get_bootstrap(self._task_bootstraps[k][bootstrap_idx])
            y = self._task_response[k].get_bootstrap(self._task_bootstraps[k][bootstrap_idx])

            MPControl.sync_processes(pref="sk_pre")

            utils.Debug.vprint('Calculating task {k} using {n}'.format(k=k, n=self._sklearn_model.__name__), level=0)
            t_beta, t_br = SKLearnRegression(x,
                                             y,
                                             self._sklearn_model,
                                             random_state=self.random_seed if self._sklearn_add_random_state else None,
                                             **self._sklearn_model_params).run()
            betas.append(t_beta)
            betas_resc.append(t_br)

        return betas, betas_resc



import numpy as np

from inferelator.postprocessing.precision_recall import RankSummaryPR
from inferelator.postprocessing import (TARGET_COLUMN, REGULATOR_COLUMN, CONFIDENCE_COLUMN,
                                        F1_COLUMN, PRECISION_COLUMN, RECALL_COLUMN)

import matplotlib
matplotlib.use('pdf')
import matplotlib.pyplot as plt


class RankSummaryF1(RankSummaryPR):
    """
    This class extends RankSumming and calculates Matthews correlation coefficient
    """

    name = "F1"
    curve_file_name = "f1_curve.pdf"

    # F1 values
    f1 = None

    # Plotter function

    @property
    def optconff1(self):
        return RankSummaryF1.calculate_opt_conf_f1(self.filtered_data)

    @property
    def maxf1(self):
        return self.calculate_opt_f1(self.filtered_data)

    def __init__(self, rankable_data, gold_standard, filter_method='keep_all_gold_standard'):
        super(RankSummaryPR, self).__init__(rankable_data, gold_standard, filter_method=filter_method)

        # Calculate the precision and recall and store them with confidence data
        self.filtered_data = self.calculate_precision_recall(self.filtered_data.copy(), transform_ties='mean')
        self.filtered_data = self.calculate_f1(self.filtered_data.copy())

        # Join the filtered F1 score onto the full confidences
        join_data = self.filtered_data.loc[:, [TARGET_COLUMN, REGULATOR_COLUMN, F1_COLUMN]]
        join_data = join_data.set_index([TARGET_COLUMN, REGULATOR_COLUMN])
        self.confidence_data = self.confidence_data.join(join_data, on=[TARGET_COLUMN, REGULATOR_COLUMN])

    def score(self):
        return self.name, self.maxf1

    def curve_dataframe(self):
        return self.filtered_data[[CONFIDENCE_COLUMN, F1_COLUMN]]

    def output_curve(self, ax=None, figsize=(6, 4)):
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize, constrained_layout=True)
            ax = fig.add_subplot(1, 1, 1)

        # Extract the recall and precision data
        curve = self.curve_dataframe()
        self.plot_f1_conf(curve[F1_COLUMN].values, curve[CONFIDENCE_COLUMN].values, self.maxf1, self.optconff1, ax,
                          num_edges=(self.confidence_data[CONFIDENCE_COLUMN] >= self.optconff1).sum())

        return ax

    @staticmethod
    def plot_f1_conf(f1, conf, optf1, optconf, ax, num_edges=None):

        num_edges = np.sum(conf >= optconf) if num_edges is None else num_edges

        # Generate a plot
        ax.plot(conf, f1)
        ax.set_xlabel('Confidence')
        ax.set_xlim(1, 0)
        ax.set_ylim(0, 1)
        ax.set_ylabel('F1')
        ax.vlines(float(optconf), 0, 1, transform=ax.get_xaxis_transform(), colors='r', linestyles='dashed')

        _msg = "max F1 = {optf1:.4f}\noptimal conf = {optconf:.4f}\nnum_edges = {n}".format(optf1=optf1,
                                                                                            optconf=optconf,
                                                                                            n=num_edges)
        ax.annotate(_msg, xy=(0.4, 0.075), xycoords='axes fraction')

        return ax

    @staticmethod
    def calculate_opt_f1(data):

        return np.nanmax(data[F1_COLUMN])

    @staticmethod
    def calculate_opt_conf_f1(data):

        return data.loc[data[F1_COLUMN] >= np.max(data[F1_COLUMN]), CONFIDENCE_COLUMN].min()

    @staticmethod
    def calculate_f1(data):

        data[F1_COLUMN] = RankSummaryF1.pr_to_f1(data[PRECISION_COLUMN], data[RECALL_COLUMN])
        return data

    @staticmethod
    def pr_to_f1(prec, recall):
        f1 = 2 * (prec * recall) / (prec + recall)
        f1[np.isnan(f1)] = 0.
        return f1



from inferelator.postprocessing.matthews_correlation import RankSummaryMCC
from inferelator.postprocessing.precision_recall import RankSummaryPR
from inferelator.postprocessing.f1_score import RankSummaryF1
from inferelator.postprocessing.model_performance import RankSummingMetric
from inferelator.postprocessing.column_names import (PRECISION_COLUMN, CONFIDENCE_COLUMN, RECALL_COLUMN, MCC_COLUMN,
                                                     F1_COLUMN, TARGET_COLUMN, REGULATOR_COLUMN)

from inferelator.utils import is_string
import os

import matplotlib
matplotlib.use('pdf')
import matplotlib.pyplot as plt


class MetricHandler(object):

    @classmethod
    def get_metric(cls, metric_ref):

        """
        This wrappers a metric reference so that strings can be used instead of python imports
        Will either return a metric class or will raise an error
        :param metric_ref: str / RankSummingMetric
            String or subclass of RankSummingMetric
        :return: RankSummingMetric
            The metadata parser that corresponds to the string, or the MetadataParser object will be passed through
        """
        if is_string(metric_ref):
            metric_ref = metric_ref.lower()
            if metric_ref == "aupr" or metric_ref == "precision-recall":
                return RankSummaryPR
            if metric_ref == "mcc" or metric_ref == "matthews correlation coefficient":
                return RankSummaryMCC
            if metric_ref == "f1" or metric_ref == "f1 score":
                return RankSummaryF1
            if metric_ref == "combined":
                return CombinedMetric
            else:
                raise ValueError("Parser {parser_str} unknown".format(parser_str=metric_ref))
        elif issubclass(metric_ref, RankSummingMetric):
            return metric_ref
        else:
            raise ValueError("Handler must be a string or a RankSummingMetric class")


class CombinedMetric(RankSummaryF1, RankSummaryPR, RankSummaryMCC):

    name = "combined"
    curve_file_name = "combined_metrics.pdf"

    def __init__(self, rankable_data, gold_standard, filter_method='keep_all_gold_standard'):

        # Call to the super __init__ to rank data
        RankSummingMetric.__init__(self, rankable_data, gold_standard, filter_method=filter_method)

        # Add Precision / Recall
        self.calculate_precision_recall(self.filtered_data)
        self.calculate_mcc(self.filtered_data)
        self.calculate_f1(self.filtered_data)

        # Join the filtered scores onto the full confidences
        join_data = self.filtered_data.loc[:, [TARGET_COLUMN, REGULATOR_COLUMN,
                                               PRECISION_COLUMN, RECALL_COLUMN, MCC_COLUMN, F1_COLUMN]]
        join_data = join_data.set_index([TARGET_COLUMN, REGULATOR_COLUMN])

        self.confidence_data = self.confidence_data.join(join_data, on=[TARGET_COLUMN, REGULATOR_COLUMN])

    def score(self):
        return self.name, self.aupr

    def all_scores(self):
        """Return a dict keyed by metric name with the metric score"""
        return dict([('AUPR', self.aupr), ('F1', self.maxf1), ('MCC', self.maxmcc)])

    @classmethod
    def all_names(cls):
        """Return a list of metric names used"""
        return ['AUPR', 'F1', 'MCC']

    def auc(self):
        return self.aupr

    def curve_dataframe(self):
        return self.filtered_data.loc[:, [CONFIDENCE_COLUMN, PRECISION_COLUMN, RECALL_COLUMN, MCC_COLUMN, F1_COLUMN]]

    def output_curve_pdf(self, output_dir, file_name=None, dpi=300, figsize=(8, 10)):

        file_name = self.curve_file_name if file_name is None else file_name

        # Create a figure
        fig, axes = plt.subplots(nrows=2, ncols=2, figsize=figsize, constrained_layout=True)

        # Draw the PR curve
        RankSummaryPR.output_curve(self, ax=axes[0, 0])

        # Add the cutoff at optimal MCC to the curve
        _ocr = self.filtered_data.loc[self.filtered_data[CONFIDENCE_COLUMN] >= self.optconfmcc, RECALL_COLUMN].max()
        _ocp = self.filtered_data.loc[self.filtered_data[CONFIDENCE_COLUMN] >= self.optconfmcc, PRECISION_COLUMN].min()
        axes[0, 0].vlines(_ocr, 0, _ocp, colors='r', linestyles='dashed')

        # Draw the MCC curve
        RankSummaryMCC.output_curve(self, ax=axes[0, 1])

        # Draw the F1 curve
        RankSummaryF1.output_curve(self, ax=axes[1, 0])

        # Draw the histogram
        self.output_histogram_edges_conf(self.filtered_data[CONFIDENCE_COLUMN].values, ax=axes[1, 1])

        # If there's a file name set, make the output file
        if file_name is not None and output_dir is not None:
            # Save the plot and close
            self.save_figure(os.path.join(output_dir, file_name), fig, dpi=dpi)

        plt.close(fig)

    @staticmethod
    def output_histogram_edges_conf(conf, ax):

        conf = conf[conf > 0.]

        ax.hist(conf, bins=50)
        ax.set_xlabel('Confidence')
        ax.set_xlim(1, 0)
        ax.set_ylabel('# Edges')
        ax.set_yscale('log')

        return ax



import pandas as pd

from inferelator.distributed.inferelator_mp import MPControl
from inferelator.utils import Debug
from inferelator.regression.base_regression import _MultitaskRegressionWorkflowMixin
from inferelator.regression.bbsr_python import BBSR, BBSRRegressionWorkflowMixin


class BBSRByTaskRegressionWorkflowMixin(_MultitaskRegressionWorkflowMixin, BBSRRegressionWorkflowMixin):
    """
    This runs BBSR regression on tasks defined by the AMUSR regression (MTL) workflow
    """

    def run_bootstrap(self, bootstrap_idx):
        betas, betas_resc = [], []

        # Select the appropriate bootstrap from each task and stash the data into X and Y
        for k in range(self._n_tasks):
            X = self._task_design[k].get_bootstrap(self._task_bootstraps[k][bootstrap_idx])
            Y = self._task_response[k].get_bootstrap(self._task_bootstraps[k][bootstrap_idx])

            # Make sure that the priors align to the expression matrix
            priors_data = self._task_priors[k].reindex(labels=self._targets, axis=0). \
                reindex(labels=self._regulators, axis=1). \
                fillna(value=0)

            if self.clr_only:
                # Create a mock prior with no information if clr_only is set
                priors_data = pd.DataFrame(0, index=priors_data.index, columns=priors_data.columns)

            MPControl.sync_processes(pref="bbsr_pre")

            Debug.vprint('Calculating MI, Background MI, and CLR Matrix', level=0)
            clr_matrix, _ = self.mi_driver().run(Y, X, return_mi=False)

            Debug.vprint('Calculating task {k} betas using BBSR'.format(k=k), level=0)
            t_beta, t_br = BBSR(X, Y, clr_matrix, priors_data,
                                prior_weight=self.prior_weight, no_prior_weight=self.no_prior_weight,
                                nS=self.bsr_feature_num).run()
            betas.append(t_beta)
            betas_resc.append(t_br)

        return betas, betas_resc



from __future__ import absolute_import, division, print_function
import time
import os
import math
import subprocess
import copy

from dask import distributed
from dask_jobqueue import SLURMCluster
from dask_jobqueue.slurm import SLURMJob

from inferelator import utils
from inferelator.utils import Validator as check
from inferelator.distributed import AbstractController
from inferelator.distributed.dask_functions import dask_map

_DEFAULT_CONDA_ACTIVATE = "source ~/.local/anaconda3/bin/activate"
_DEFAULT_NUM_JOBS = 1
_DEFAULT_THREADS_PER_WORKER = 1
_DEFAULT_WORKERS_PER_JOB = 20
_DEFAULT_MEM_PER_JOB = '62GB'
_DEFAULT_INTERFACE = 'ib0'
_DEFAULT_WALLTIME = '1:00:00'

_DEFAULT_ENV_EXTRA = ['module purge']

_THREAD_CONTROL_ENV = ['export MKL_NUM_THREADS={t}',
                       'export OPENBLAS_NUM_THREADS={t}',
                       'export NUMEXPR_NUM_THREADS={t}']

_DEFAULT_CONTROLLER_EXTRA = ['--nodes 1', '--ntasks-per-node 1']

_DEFAULT_ADAPTIVE_INTERVAL = "1s"

_KNOWN_CONFIG = {"prince": {"_job_n_workers": 20,
                            "_job_mem": "62GB",
                            "_job_time": "48:00:00",
                            "_interface": "ib0",
                            "_job_extra_env_commands": copy.copy(_DEFAULT_ENV_EXTRA)
                            },

                 "rusty_ccb": {"_job_n_workers": 28,
                               "_num_local_workers": 25,
                               "_job_mem": "498GB",
                               "_job_time": "48:00:00",
                               "_interface": "ib0",
                               "_queue": "ccb",
                               "_job_extra_env_commands": copy.copy(_DEFAULT_ENV_EXTRA)
                               },

                 "rusty_preempt": {"_job_n_workers": 40,
                                   "_num_local_workers": 35,
                                   "_job_mem": "766GB",
                                   "_job_time": "48:00:00",
                                   "_interface": "ib0",
                                   "_queue": "preempt",
                                   "_job_extra_env_commands": copy.copy(_DEFAULT_ENV_EXTRA),
                                   "_job_slurm_commands": copy.copy(_DEFAULT_CONTROLLER_EXTRA) + ["--qos=preempt",
                                                                                                  "--constraint=info"]
                                   },

                 "rusty_rome": {"_job_n_workers": 64,
                                "_job_n_threads": 2,
                                "_num_local_workers": 60,
                                "_job_mem": "990GB",
                                "_job_time": "24:00:00",
                                "_interface": "ib0",
                                "_queue": "ccb",
                                "_job_extra_env_commands": copy.copy(_DEFAULT_ENV_EXTRA),
                                "_job_slurm_commands": copy.copy(_DEFAULT_CONTROLLER_EXTRA) + ["--constraint=rome"]
                                },
                 }


_DEFAULT_LOCAL_WORKER_COMMAND = "dask-worker {a} --nprocs {p} --nthreads {t} --memory-limit 0 --local-directory {d}"

try:
    _DEFAULT_LOCAL_DIR = os.environ['TMPDIR']
except KeyError:
    _DEFAULT_LOCAL_DIR = 'dask-worker-space'

try:
    _DEFAULT_SLURM_ID = os.environ['SLURM_JOB_ID']
except KeyError:
    _DEFAULT_SLURM_ID = "1"


# This is the worst thing I've ever written
def memory_limit_0(command_template):
    return "--".join(["memory-limit 0 " if c.startswith("memory-limit") else c for c in command_template.split("--")])


class SLURMJobNoMemLimit(SLURMJob):
    def __init__(self, *args, **kwargs):
        super(SLURMJobNoMemLimit, self).__init__(*args, **kwargs)
        self._command_template = memory_limit_0(self._command_template)


class DaskHPCClusterController(AbstractController):
    """
    The DaskHPCClusterController launches a HPC cluster and connects as a client. By default it uses the SLURM workload
    manager, but other workload managers may be used.
    #TODO: test drop-in replacement of other managers

    Many of the cluster-specific options are taken from class variables that can be set prior to calling .connect().

    The map functionality is deliberately not implemented; dask-specific multiprocessing functions are used instead
    """
    _controller_name = "dask-cluster"
    _controller_dask = True

    is_master = True
    client = None

    # Cluster Controller
    _cluster_controller_class = SLURMCluster

    # The dask cluster object
    _local_cluster = None
    _tracker = None

    # Should any local workers be started on this node
    _num_local_workers = 0
    _runaway_protection = 3

    # SLURM specific variables
    _queue = None
    _project = None
    _interface = _DEFAULT_INTERFACE
    _local_directory = _DEFAULT_LOCAL_DIR

    # Job variables
    _job_n = _DEFAULT_NUM_JOBS
    _job_n_workers = _DEFAULT_WORKERS_PER_JOB
    _worker_n_threads = _DEFAULT_THREADS_PER_WORKER
    _job_mem = _DEFAULT_MEM_PER_JOB
    _job_time = _DEFAULT_WALLTIME
    _job_slurm_commands = copy.copy(_DEFAULT_CONTROLLER_EXTRA)
    _job_extra_env_commands = copy.copy(_DEFAULT_ENV_EXTRA)
    _job_threading_commands = copy.copy(_THREAD_CONTROL_ENV)

    @classmethod
    def connect(cls, *args, **kwargs):
        """
        Setup slurm cluster
        """

        # Create a slurm cluster with all the various class settings
        cls._local_cluster = cls._cluster_controller_class(queue=cls._queue,
                                                           project=cls._project,
                                                           interface=cls._interface,
                                                           walltime=cls._job_time,
                                                           job_cpu=cls._job_n_workers * cls._worker_n_threads,
                                                           cores=cls._job_n_workers * cls._worker_n_threads,
                                                           processes=cls._job_n_workers,
                                                           threads = cls._worker_n_threads,
                                                           job_mem=cls._job_mem,
                                                           env_extra=cls._config_env(),
                                                           local_directory=cls._local_directory,
                                                           memory=cls._job_mem,
                                                           job_extra=cls._job_slurm_commands,
                                                           job_cls=SLURMJobNoMemLimit)

        cls.client = distributed.Client(cls._local_cluster, direct_to_workers=True)

        cls._add_local_node_workers(cls._num_local_workers)
        cls._tracker = WorkerTracker()

        return True

    @classmethod
    def shutdown(cls):
        cls.client.close()
        cls._local_cluster.close()

    @classmethod
    def map(cls, func, *args, **kwargs):
        return dask_map(func, *args, **kwargs)

    @classmethod
    def use_default_configuration(cls, known_config, n_jobs=1):
        """
        Load a known default cluster configuration

        :param known_config: A string with a valid known cluster configuration.
        Currently implemented are "prince" (NYU) and "rusty_ccb" (Simons Foundation; Center for Computational Biology)
        :type known_config: str
        :param n_jobs: Number of jobs to start with this configuration
        :type n_jobs: int
        """

        known_config = known_config.lower()
        if known_config in _KNOWN_CONFIG:
            for k, v in _KNOWN_CONFIG[known_config].items():
                setattr(cls, k, v)
            cls._job_n = n_jobs
        else:
            msg = "Configuration {k} is unknown".format(k=known_config)
            raise ValueError(msg)

        utils.Debug.vprint(cls._config_str(), level=1)

    @classmethod
    def set_job_size_params(cls, n_jobs=None, n_cores_per_job=None, mem_per_job=None, walltime=None,
                            n_workers_per_job=None, n_threads_per_worker=None):
        """
        Set the job size parameters

        :param n_jobs: The number of jobs to start. Each job will have separate workers and memory allocated.
        This is in addition to the main (scheduler) job.
        For slurm, each job will be set with "#SBATCH --nodes=1"
        :type n_jobs: int
        :param n_cores_per_job: The number of workers to start per job.
        For SLURM, this is setting "#SBATCH --cpus-per-task"
        :type n_cores_per_job: int
        :param mem_per_job: The amount of memory to allocate per job.
        For SLURM, this is setting "#SBATCH --mem"
        :type mem_per_job: str, int
        :param walltime: The time limit per worker job.
        For SLURM, this is setting #SBATCH --time
        :type walltime: str
        :param n_workers_per_job: The number of worker jobs to start
        SLURM will allocate n_workers_per_job * n_threads_per_worker cores per job.
        :type n_workers_per_job: int
        :param n_threads_per_worker: The number of threads to give each worker job.
        SLURM will allocate n_workers_per_job * n_threads_per_worker cores per job.

        """

        check.argument_integer(n_jobs, allow_none=True)
        check.argument_integer(n_cores_per_job, allow_none=True)
        check.argument_integer(n_threads_per_worker, allow_none=True)

        cls._job_n = n_jobs if n_jobs is not None else cls._job_n
        cls._job_n_workers = n_cores_per_job if n_cores_per_job is not None else cls._job_n_workers
        cls._job_mem = mem_per_job if mem_per_job is not None else cls._job_mem
        cls._job_time = walltime if walltime is not None else cls._job_time
        cls._worker_n_threads = n_threads_per_worker if n_threads_per_worker is not None else cls._worker_n_threads
        cls._job_n_workers = n_workers_per_job if n_workers_per_job is not None else cls._job_n_workers

    @classmethod
    def set_cluster_params(cls, queue=None, project=None, interface=None, local_workers=None):
        """
        Set parameters which are specific to the HPC environment

        :param queue: The name of the SLURM partition to use. If None, do not specify a partition. Defaults to None.
        :type queue: str
        :param project: The name of the project for each job. If None, do not specify a project. Defaults to None.
        :type project: str
        :param interface: A string that identifies the network interface to use.
        Possible options may include 'eth0' or 'ib0'.
        :type interface: str
        :param local_workers: The number of local workers to start on the same node as the main scheduler
        :type local_workers: int
        """

        cls._queue = queue if queue is not None else cls._queue
        cls._project = project if project is not None else cls._project
        cls._interface = interface if interface is not None else cls._interface
        cls._num_local_workers = local_workers if local_workers is not None else cls._num_local_workers

    @classmethod
    def set_processes(cls, process_count):
        """
        Set the number of dask workers to use
        :param process_count: int
        :return:
        """
        check.argument_integer(process_count, low=1)
        cls._job_n = math.ceil(process_count / cls._job_n_workers)

        utils.Debug.vprint("Using `set_processes` is not advised for the DASK CLUSTER configuration", level=0)
        utils.Debug.vprint("Using `set_job_size_params` is highly preferred", level=0)
        utils.Debug.vprint("Configured {n} jobs with {w} workers per job".format(n=cls._job_n, w=cls._job_n_workers),
                           level=0)

    @classmethod
    def sync_processes(cls, *args, **kwargs):
        """
        This is a thing for KVS. Just return True.
        """
        return True

    @classmethod
    def add_worker_env_line(cls, line):
        """
        Add a line to the worker environment declaration
        This gets put into the sbatch script for workers
        This can be used to load modules, activate conda, etc
        """

        cls._job_extra_env_commands.append(line)

    @classmethod
    def add_slurm_command_line(cls, line):
        """
        Add a line to the command block for SLURM.
        This will be prepended with the "#SBATCH " command
        """

        cls._job_slurm_commands.append(line)

    @classmethod
    def add_worker_conda(cls, cmd=None, env=None):
        """
        Add a line to activate a conda environment for workers

        :param cmd: A shell command which will activate a conda environment
        Defaults to "source ~/.local/anaconda3/bin/activate base"
        :type cmd: str
        :param env: The environment to activate. This will be used with the $CONDA_PREFIX variables if no argument
        is provided for `cmd`.
        :type env: str
        """

        if cmd is None:
            try:
                conda = os.environ['CONDA_PREFIX_1'] if 'CONDA_PREFIX_1' in os.environ else os.environ['CONDA_PREFIX']
                cmd = os.path.join(conda, "bin", "activate")
                cmd = "source " + cmd + " " + (env if env is not None else "")
                cls.add_worker_env_line(cmd)
                return True
            except KeyError:
                raise ValueError("Unable to set conda environment implicitly; pass explicit `cmd=` argument instead")
        elif env is not None:
            utils.Debug.vprint("The `env` argument to `add_worker_conda` is ignored when `cmd` is passed", level=0)

        cls.add_worker_env_line(cmd)
        return True

    @classmethod
    def is_dask(cls):
        """
        Block when something asks if this is a dask function until the workers are alive
        """

        cls.check_cluster_state()
        return True

    @classmethod
    def check_cluster_state(cls):
        """
        Check to make sure that workers have been provisioned. Sleep until workers become available.
        """

        cls._scale_jobs()

        sleep_time = 0
        while not cls._local_cluster.observed:
            time.sleep(1)
            if sleep_time % 60 == 0:
                utils.Debug.vprint("Awaiting workers ({t} seconds elapsed)".format(t=sleep_time), level=0)
            sleep_time += 1

    @classmethod
    def _config_str(cls):
        status = "\n".join(["Dask cluster: Allocated {n} jobs ({w} workers with {m} memory per job)",
                            "SLURM: -p {q}, -A {p}, " + ", ".join(cls._job_slurm_commands),
                            "ENV: " + "\n\t".join(cls._job_extra_env_commands)]) + "\n"
        return status.format(n=cls._job_n, w=cls._job_n_workers, m=cls._job_mem, q=cls._queue, p=cls._project)

    @classmethod
    def _config_env(cls):
        return [s.format(t=cls._worker_n_threads) for s in cls._job_threading_commands] + cls._job_extra_env_commands

    @classmethod
    def _scale_jobs(cls):
        """
        Update the worker tracker. If an entire slurm job is dead, start a new one to replace it.
        """
        cls._tracker.update_lists(cls._local_cluster.observed, cls._local_cluster.worker_spec)

        new_jobs = cls._job_n + cls._tracker.num_dead

        if cls._runaway_protection is not None and new_jobs > cls._runaway_protection * cls._job_n:
            raise RuntimeError("Aborting excessive worker startups / Protecting against runaway job queueing")
        elif new_jobs > len(cls._local_cluster.worker_spec):
            cls._local_cluster.scale(jobs=new_jobs)

    @classmethod
    def _add_local_node_workers(cls, num_workers):
        """
        Start workers on the local node with the scheduler & client

        :param num_workers: The number of workers to start on this node
        :type num_workers: int
        """
        check.argument_integer(num_workers, low=0, allow_none=True)

        if num_workers is not None and num_workers > 0:
            cmd = _DEFAULT_LOCAL_WORKER_COMMAND.format(p=num_workers,
                                                       t=cls._worker_n_threads,
                                                       a=cls._local_cluster.scheduler_address,
                                                       d=cls._local_directory)
            out_handle = open("slurm-{i}.out".format(i=_DEFAULT_SLURM_ID), mode="w")
            subprocess.Popen(cmd, shell=True, stdout=out_handle, stderr=out_handle)


class WorkerTracker:
    """
    Keep track of which workers have been started but are now gone
    Workers with no live procs will be assumed dead
    """

    def __init__(self):
        self._live_workers = set()
        self._dead_workers = set()

        self._dead_cluster_job = set()

    def update_lists(self, current_alive, worker_spec):
        self._dead_workers.update(self._live_workers.difference(current_alive))
        self._live_workers.update(current_alive)

        for k, v in worker_spec.items():
            if k in self._dead_cluster_job:
                pass

            workers_in_spec = set(str(k) + g for g in v['group'])
            if workers_in_spec.issubset(self._dead_workers):
                self._dead_cluster_job.add(k)

    @property
    def num_dead(self):
        return len(self._dead_cluster_job)



# Set threading control variables if they're not already set

import os
os.environ["MKL_NUM_THREADS"] = os.environ.get("MKL_NUM_THREADS", "1")
os.environ["NUMEXPR_NUM_THREADS"] = os.environ.get("NUMEXPR_NUM_THREADS", "1")
os.environ["OMP_NUM_THREADS"] = os.environ.get("OMP_NUM_THREADS", "1")
os.environ["OPENBLAS_NUM_THREADS"] = os.environ.get("OPENBLAS_NUM_THREADS", "1")

from inferelator.workflow import inferelator_workflow
from inferelator.crossvalidation_workflow import CrossValidationManager
from inferelator.utils import inferelator_verbose_level
from inferelator.distributed.inferelator_mp import MPControl




"""
This file sets default constants that are used throughout the package.
Objects set as defaults are set in classes (so that we don't have to import anything here).
Mostly I did this so I could easily reuse defaults and change them to match my data structure.
Don't look at me like that.
"""

"""Default Environmental Variable Lookup"""

# This is a dict, keyed by the class setattr variable name, of tuples (env name, coercion function, default value)
SBATCH_VARS = dict(output_dir=('RUNDIR', str, None),
                   input_dir=('DATADIR', str, None),
                   rank=('SLURM_PROCID', int, 0),
                   cores=('SLURM_NTASKS_PER_NODE', int, 1),
                   tasks=('SLURM_NTASKS', int, 1),
                   node=('SLURM_NODEID', int, 0),
                   num_nodes=('SLURM_JOB_NUM_NODES', int, 1))

SBATCH_VARS_FOR_KVS = ["rank", "cores", "tasks", "node", "num_nodes"]
SBATCH_VARS_FOR_WORKFLOW = ["output_dir", "input_dir"]

"""Default Data File Settings"""

DEFAULT_PD_INPUT_SETTINGS = dict(sep="\t")
DEFAULT_EXPRESSION_FILE = "expression.tsv"
DEFAULT_TFNAMES_FILE = "tf_names.tsv"
DEFAULT_METADATA_FILE = "meta_data.tsv"
DEFAULT_PRIORS_FILE = "gold_standard.tsv"
DEFAULT_GOLDSTANDARD_FILE = "gold_standard.tsv"

"""Default TFAWorkflow Parameters"""
DEFAULT_DELTMIN = 0
DEFAULT_DELTMAX = 120
DEFAULT_TAU = 45
DEFAULT_GS_FILTER_METHOD = 'keep_all_gold_standard'

"""Defaults For Regression"""
# Default number of predictors to include in the model
DEFAULT_nS = 10

# Default weight for priors & Non-priors
# If prior_weight is the same as no_prior_weight:
#   Priors will be included in the pp matrix before the number of predictors is reduced to nS
#   They won't get special treatment in the model though
DEFAULT_prior_weight = 1
DEFAULT_no_prior_weight = 1

# Throw away the priors which have a CLR that is 0 before the number of predictors is reduced by BIC
DEFAULT_filter_priors_for_clr = False



"""
Construct inferelator workflows from preprocessing, postprocessing, and regression modules
"""
from __future__ import unicode_literals, print_function

import datetime
import inspect
import os
import warnings
import copy

import numpy as np
import pandas as pd

from inferelator.utils import (Debug, InferelatorDataLoader, DEFAULT_PANDAS_TSV_SETTINGS, slurm_envs, is_string,
                               DotProduct)
from inferelator.distributed.inferelator_mp import MPControl
from inferelator.preprocessing import ManagePriors, make_data_noisy
from inferelator.regression.base_regression import _RegressionWorkflowMixin
from inferelator.postprocessing import ResultsProcessor, InferelatorResults

SBATCH_VARS_FOR_WORKFLOW = ["output_dir", "input_dir"]

_TENX = "tenx"
_TSV = "tsv"
_H5AD = "h5ad"
_HDF5 = "hdf5"
_MTX = "mtx"


class WorkflowBaseLoader(object):
    """
    WorkflowBaseLoader is the class to load raw data. It does no processing; it only takes data from files.
    """
    # Paths to the input and output locations
    input_dir = None
    output_dir = None

    # File names for each of the data files which can be used in the inference workflow
    expression_matrix_file = None
    tf_names_file = None
    meta_data_file = None
    priors_file = None
    gold_standard_file = None

    # The expression file type
    _expression_loader = _TSV
    _h5_layer = None

    # Metadata handler
    metadata_handler = "branching"

    # Gene list & associated metadata
    gene_names_file = None
    gene_metadata_file = None
    gene_list_index = None

    # Loaded data structures [G: Genes, K: Predictors, N: Conditions]
    tf_names = None  # tf_names list [K]
    gene_names = None  # gene_names list [G]
    priors_data = None  # priors data dataframe [G x K]
    gold_standard = None  # gold standard dataframe [G x K]

    # Loaded experimental data
    data = None  # InferelatorData [N x G]

    # Calculated data structures
    design = None  # InferelatorData [N x K]
    response = None  # InferelatorData [N x G]

    # Flag to identify orientation of the expression matrix (True for samples x genes & False for genes x samples)
    expression_matrix_columns_are_genes = True  # bool

    # Flag to extract metadata from specific columns of the expression matrix instead of a separate file
    extract_metadata_from_expression_matrix = False  # bool
    expression_matrix_metadata = None  # str

    # Flag to indicate the inferelator should be run without existing network data
    use_no_prior = False  # bool
    use_no_gold_standard = False  # bool

    # Settings that will be used by pd.read_table to import data files
    _file_format_settings = None

    @property
    def _num_obs(self):
        """
        Get the number of observations N
        :return: Number of observations / samples
        :rtype: int
        """
        if self.response is not None:
            return self.response.num_obs
        elif self.data is not None:
            return self.data.num_obs
        else:
            return None

    @property
    def _num_tfs(self):
        """
        Get the number of regulators K
        :return: Number of regulators
        :rtype: int
        """
        if self.design is not None:
            return self.design.num_genes
        elif self.tf_names is not None:
            return len(self.tf_names)
        else:
            return None

    @property
    def _num_genes(self):
        """
        Get the number of genes G
        :return: Number of genes
        :rtype: int
        """
        if self.response is not None:
            return self.response.num_genes
        elif self.data is not None:
            return self.data.num_genes
        else:
            return None

    def __init__(self):
        if self._file_format_settings is None:
            self._file_format_settings = dict()

    def set_file_paths(self, input_dir=None, output_dir=None, expression_matrix_file=None, tf_names_file=None,
                       meta_data_file=None, priors_file=None, gold_standard_file=None, gene_metadata_file=None,
                       gene_names_file=None):
        """
        Set the file paths necessary for the inferelator to run

        :param input_dir: A path containing the input files
        :type input_dir: str
        :param output_dir: A path to put the output files
        :type output_dir: str, optional
        :param expression_matrix_file: Path to the expression data
            If set here, this expression file will be assumed to be a TSV file.
            Use set_expression_file() for other file types
        :type expression_matrix_file: str
        :param meta_data_file: Path to the meta data TSV file
        :type meta_data_file: str, optional
        :param tf_names_file: Path to a list of regulator names to include in the model
        :type tf_names_file: str
        :param priors_file: Path to a prior data file TSV file [Genes x Regulators]
        :type priors_file: str
        :param gold_standard_file: Path to a gold standard data TSV file [Genes x Regulators]
        :type gold_standard_file: str
        :param gene_metadata_file: Path to a genes annotation file
        :type gene_metadata_file: str, optional
        :param gene_names_file: Path to a list of genes to include in the model (optional)
        :type gene_names_file: str, optional
        """

        self._set_with_warning("input_dir", input_dir)
        self._set_with_warning("output_dir", output_dir)
        self._set_file_name("expression_matrix_file", expression_matrix_file)
        self._set_file_name("tf_names_file", tf_names_file)
        self._set_file_name("meta_data_file", meta_data_file)
        self._set_file_name("priors_file", priors_file)
        self._set_file_name("gold_standard_file", gold_standard_file)
        self._set_file_name("gene_metadata_file", gene_metadata_file)
        self._set_file_name("gene_names_file", gene_names_file)

        if expression_matrix_file is not None:
            self._expression_loader = _TSV
            if ".tsv" not in expression_matrix_file:
                msg = "`set_file_paths` assumes data is in a TSV. Use `set_expression_file` for other formats"
                warnings.warn(msg)

    def set_expression_file(self, tsv=None, hdf5=None, h5ad=None, tenx_path=None, mtx=None, mtx_barcode=None,
                            mtx_feature=None, h5_layer=None):
        """
        Set the type of expression data file. Current loaders include TSV, hdf5, h5ad (AnnData), and MTX sparse files.
        Only one of these loaders can be used; passing arguments for multiple loaders will raise a ValueError.

        :param tsv: A path to a TSV (or tsv.gz) file which can be loaded by pandas.read_csv()
        :type tsv: str, optional
        :param hdf5: A path to a hdf5 file which can be loaded by pandas.HDFStore
        :type hdf5: str, optional
        :param h5ad: A path to an AnnData hd5 file
        :type h5ad: str, optional
        :param tenx_path: A path to the folder containing the 10x mtx, barcode, and feature files
        :type tenx_path: Path, optional
        :param mtx: A path to an mtx file
        :type mtx: str, optional
        :param mtx_barcode: A path to a list of observation names (i.e. barcodes, etc) for the mtx file
        :type mtx_barcode: str, optional
        :param mtx_feature: A path to a list of gene names for the mtx file
        :type mtx_feature: str, optional
        :param h5_layer: The layer (in an AnnData h5) or the store key (in an hdf5) file to use.
            Defaults to using the first key.
        :type h5_layer: str, optional
        """

        nones = [tsv is None, hdf5 is None, h5ad is None, tenx_path is None, mtx is None]

        if all(nones):
            Debug.vprint("No file provided", level=0)
        elif sum(nones) != (len(nones) - 1):
            raise ValueError("Only one type of input expression file can be set")

        if tsv is not None:
            self._set_file_name("expression_matrix_file", tsv)
            self._expression_loader = _TSV
        elif hdf5 is not None:
            self._set_file_name("expression_matrix_file", hdf5)
            self._expression_loader = _HDF5
            self._h5_layer = h5_layer
        elif h5ad is not None:
            self._set_file_name("expression_matrix_file", h5ad)
            self._expression_loader = _H5AD
            self._h5_layer = h5_layer
        elif mtx is not None:
            self._check_file_exists(mtx)
            self._check_file_exists(mtx_barcode)
            self._check_file_exists(mtx_feature)
            self.expression_matrix_file = (mtx, mtx_barcode, mtx_feature)
            self._expression_loader = _MTX
        elif tenx_path is not None:
            self.expression_matrix_file = tenx_path
            self._expression_loader = _TENX

    def set_file_properties(self, extract_metadata_from_expression_matrix=None, expression_matrix_metadata=None,
                            expression_matrix_columns_are_genes=None, gene_list_index=None, metadata_handler=None):
        """
        Set properties associated with the input data files

        :param extract_metadata_from_expression_matrix: A boolean flag that should be set to True if there is
            non-expression data in the expression matrix. If True, `expression_matrix_metadata` must be provided.
            Defaults to False.
        :type extract_metadata_from_expression_matrix: bool, optional
        :param expression_matrix_metadata: A list of columns which, if provided, will be removed from
            the expression matrix file and kept as metadata.
            Defaults to None.
        :type expression_matrix_metadata: list(str), optional
        :param expression_matrix_columns_are_genes: A boolean flag indicating the orientation of the expression matrix.
            False reads the expression matrix as genes on rows, samples on columns.
            True reads the expression matrix as samples on rows, genes on columns.
            Defaults to False.
        :type expression_matrix_columns_are_genes: bool, optional
        :param gene_list_index: The column name in the gene metadata file which corresponds to the gene labels in the
            expression and prior data files.
            Defaults to None.
            Must be provided if `gene_metadata_file` was provided to `set_file_paths()`.
        :type gene_list_index: str, optional
        :param metadata_handler: A string which identifies the specific metadata parsing method to use. Options include
            "branching" or "nonbranching". Defaults to "branching".
        :type metadata_handler: str
        """

        if extract_metadata_from_expression_matrix is not None:
            warnings.warn("Set expression_matrix_metadata to extract columns", DeprecationWarning)

        self._set_without_warning("expression_matrix_columns_are_genes", expression_matrix_columns_are_genes)

        self._set_with_warning("expression_matrix_metadata", expression_matrix_metadata)
        self._set_with_warning("gene_list_index", gene_list_index)
        self._set_with_warning("metadata_handler", metadata_handler)

    def set_network_data_flags(self, use_no_prior=None, use_no_gold_standard=None):
        """
        Set flags to skip using existing network data. Note that these flags will be ignored if network data is
        provided

        :param use_no_prior: Flag to indicate the inferelator should be run without existing prior data.
            Will create a mock prior with no information. Highly inadvisable. Defaults to False
        :type use_no_prior: bool
        :param use_no_gold_standard: Flag to indicate the inferelator should be run without existing gold standard data.
            Will create a mock gold standard with no information. Highly inadvisable. Defaults to False
        :type use_no_gold_standard: bool
        """

        warnings.warn("Omitting prior network data is not recommended. Use at your own risk.")

        self._set_without_warning("use_no_prior", use_no_prior)
        self._set_without_warning("use_no_gold_standard", use_no_gold_standard)

    def set_file_loading_arguments(self, file_name, **kwargs):
        """
        Update the settings for a given file name. By default we assume all files can be read in as TSV files.
        Any arguments provided here will be passed to `pandas.read_csv()` for the file name provided.

        `set_file_loading_arguments('expression_matrix_file', sep=",")` will read the expression_matrix_file as a CSV.

        :param file_name: The name of the variable containing the file name (from `set_file_properties`)
        :type file_name: str
        :param kwargs: Arguments to be passed to `pandas.read_csv()`
        """

        # Check and see if file_name is actually an object attribute holding a file name. Use that if so.
        file_name = self._get_file_name_from_attribute(file_name)
        if file_name is None:
            return

        self._file_format_settings[file_name].update(kwargs)
        self.print_file_loading_arguments(file_name)

    def print_file_loading_arguments(self, file_name):
        """
        Print the settings that will be used to load a given file name.

        :param file_name: The name of the variable containing the file name (from `set_file_properties`)
        :type file_name: str
        """

        # Check and see if file_name is actually an object attribute holding a file name. Use that if so.
        file_name = self._get_file_name_from_attribute(file_name)
        if file_name is None:
            return

        msg = "File {f} has the following settings:".format(f=file_name)
        msg += "\n\t".join([str(k) + " = " + str(v) for k, v in self._file_format_settings[file_name].items()])
        Debug.vprint(msg, level=0)

    def _get_file_name_from_attribute(self, file_name):
        """
        Check and see if a file name is an object attribute that holds a file namee
        :param file_name: str
        :return file_name: str
        """
        # Check and see if file_name is actually an object attribute holding a file name. Use that if so.
        if file_name not in self._file_format_settings:
            if hasattr(self, file_name) and getattr(self, file_name) in self._file_format_settings:
                file_name = getattr(self, file_name)
            else:
                Debug.vprint("File {f} is unknown".format(f=file_name), level=0)
                return None
        return file_name

    def _set_file_name(self, attr_name, file_name):
        """
        Set a file name. Create a set of default loading parameters and store them in _file_format_settings.
        Also print a warning if the file doesn't exist
        :param attr_name: str
        :param file_name: str
        """

        if file_name is None:
            return

        if file_name not in self._file_format_settings:
            self._file_format_settings[file_name] = copy.copy(DEFAULT_PANDAS_TSV_SETTINGS)

        self._check_file_exists(file_name)
        self._set_with_warning(attr_name, file_name)

    def _check_file_exists(self, file_name):
        """
        Print a warning if a file doesn't exist
        :param file_name: str
        """

        if file_name is not None and not os.path.isfile(self.input_path(file_name)):
            Debug.vprint("File {f} does not exist".format(f=file_name), level=0)

    def _set_without_warning(self, attr_name, value):
        """
        Set an attribute name.
        :param attr_name: str
        :param value:
        """

        if value is None:
            return

        setattr(self, attr_name, value)

    def _set_with_warning(self, attr_name, value):
        """
        Set an attribute name. Warn if it's already not None
        :param attr_name: str
        :param value:
        """

        if value is None:
            return

        current_value = getattr(self, attr_name)
        if current_value is not None:
            _msg = "Setting {a}: replacing value {vo} with value {vn}".format(a=attr_name, vo=current_value, vn=value)
            warnings.warn(_msg)

        setattr(self, attr_name, value)

    def get_data(self):
        """
        Read data files in to data structures.
        """

        self.read_expression()
        self.read_tfs()
        self.read_priors()

        # Validate that necessary input settings exist
        self.validate_data()

    def append_to_path(self, var_name, to_append):
        """
        Add a string to an existing path variable

        :param var_name: The name of the path variable (`input_dir` or `output_dir`)
        :type var_name: str
        :param to_append: The path to join to the end of the existing path variable
        :type to_append: str

        """
        path = getattr(self, var_name, None)
        if path is None:
            raise ValueError("Cannot append {to_append} to {var_name} (Which is None)".format(to_append=to_append,
                                                                                              var_name=var_name))
        setattr(self, var_name, os.path.join(path, to_append))

    def read_expression(self, expression_matrix_file=None, meta_data_file=None, gene_data_file=None):
        """
        Read expression matrix file into an InferelatorData object
        """
        expression_file = expression_matrix_file if expression_matrix_file is not None else self.expression_matrix_file
        meta_data_file = meta_data_file if meta_data_file is not None else self.meta_data_file
        gene_data_file = gene_data_file if gene_data_file is not None else self.gene_metadata_file

        loader = InferelatorDataLoader(input_dir=self.input_dir, file_format_settings=self._file_format_settings)

        if self._expression_loader == _H5AD:
            self.data = loader.load_data_h5ad(expression_file,
                                              use_layer=self._h5_layer,
                                              meta_data_file=meta_data_file,
                                              meta_data_handler=self.metadata_handler,
                                              gene_data_file=gene_data_file,
                                              gene_name_column=self.gene_list_index)

        elif self._expression_loader == _TSV:
            self.data = loader.load_data_tsv(expression_file,
                                             transpose_expression_data=not self.expression_matrix_columns_are_genes,
                                             expression_matrix_metadata=self.expression_matrix_metadata,
                                             meta_data_file=meta_data_file,
                                             meta_data_handler=self.metadata_handler,
                                             gene_data_file=gene_data_file,
                                             gene_name_column=self.gene_list_index)

        elif self._expression_loader == _MTX:
            self.data = loader.load_data_mtx(expression_file[0],
                                             mtx_feature=expression_file[2],
                                             mtx_obs=expression_file[1],
                                             meta_data_file=meta_data_file,
                                             meta_data_handler=self.metadata_handler,
                                             gene_data_file=gene_data_file,
                                             gene_name_column=self.gene_list_index)

        elif self._expression_loader == _TENX:
            self.data = loader.load_data_tenx(expression_file,
                                              meta_data_file=meta_data_file,
                                              meta_data_handler=self.metadata_handler,
                                              gene_data_file=gene_data_file,
                                              gene_name_column=self.gene_list_index)

        elif self._expression_loader == _HDF5:
            self.data = loader.load_data_hdf5(expression_file,
                                              transpose_expression_data=not self.expression_matrix_columns_are_genes,
                                              use_layer=self._h5_layer,
                                              meta_data_file=meta_data_file,
                                              meta_data_handler=self.metadata_handler,
                                              gene_data_file=gene_data_file,
                                              gene_name_column=self.gene_list_index)

        self.data.name = "Expression"

    def read_tfs(self, file=None):
        """
        Read tf names file into tf_names
        """

        # Load the class variable if no file is passed
        file = self.tf_names_file if file is None else file

        if file is not None:
            Debug.vprint("Loading TF feature names from file {file}".format(file=file), level=1)
            # Read in a dataframe with no header or index
            loader = InferelatorDataLoader(input_dir=self.input_dir, file_format_settings=self._file_format_settings)
            tfs = loader.input_dataframe(file, header=None, index_col=None)

            # Cast the dataframe into a list
            assert tfs.shape[1] == 1
            self.tf_names = tfs.values.flatten().tolist()

    def read_genes(self, file=None):
        """
        Read tf names file into tf_names
        """

        # Load the class variable if no file is passed
        file = self.gene_names_file if file is None else file

        if file is not None:
            Debug.vprint("Loading TF feature names from file {file}".format(file=file), level=1)
            # Read in a dataframe with no header or index
            loader = InferelatorDataLoader(input_dir=self.input_dir, file_format_settings=self._file_format_settings)
            genes = loader.input_dataframe(file, header=None, index_col=None)

            # Cast the dataframe into a list
            assert genes.shape[1] == 1
            self.gene_names = genes.values.flatten().tolist()

    def read_priors(self, priors_file=None, gold_standard_file=None):
        """
        Read in the priors and gold standard files
        """
        priors_file = priors_file if priors_file is not None else self.priors_file
        gold_standard_file = gold_standard_file if gold_standard_file is not None else self.gold_standard_file
        loader = InferelatorDataLoader(input_dir=self.input_dir, file_format_settings=self._file_format_settings)

        if priors_file is not None:
            Debug.vprint("Loading prior data from file {file}".format(file=priors_file), level=1)
            self.priors_data = loader.input_dataframe(priors_file)
            self.loaded_file_info("Priors data", self.priors_data)

        if gold_standard_file is not None:
            Debug.vprint("Loading gold_standard data from file {file}".format(file=gold_standard_file), level=1)
            self.gold_standard = loader.input_dataframe(gold_standard_file)
            self.loaded_file_info("Gold standard", self.gold_standard)

    def validate_data(self):
        """
        Make sure that the data that's loaded is acceptable
        """

        # Create a null prior if the flag is set
        if self.use_no_prior and self.priors_data is not None:
            warnings.warn("The use_no_prior flag will be ignored because prior data exists")
        elif self.use_no_prior:
            Debug.vprint("A null prior is has been created", level=0)
            self.priors_data = self._create_null_prior(self.data.gene_names, self.tf_names)

        # Create a null gold standard if the flag is set
        if self.use_no_gold_standard and self.gold_standard is not None:
            warnings.warn("The use_no_gold_standard flag will be ignored because gold standard data exists")
        elif self.use_no_gold_standard:
            Debug.vprint("A null gold standard has been created", level=0)
            self.gold_standard = self._create_null_prior(self.data.gene_names, self.tf_names)

        # Validate that some network information exists and has been loaded
        if self.priors_data is None and self.gold_standard is None:
            raise ValueError("No gold standard or priors have been provided")

    def input_path(self, filename):
        """
        Join filename to input_dir

        :param filename: Path to some file that needs to be attached to the input path
        :type filename: str
        :return: File joined to input_dir instance variable
        :rtype: str
        """

        return InferelatorDataLoader.filename_path_join(self.input_dir, filename)

    def output_path(self, filename):
        """
        Join filename to output_dir

        :param filename: Path to some file that needs to be attached to the output path
        :type filename: str
        :return: File joined to output_dir instance variable
        :rtype: str
        """
        return InferelatorDataLoader.filename_path_join(self.output_dir, filename)

    def load_data_and_save_h5ad(self, output_file_name, to_sparse=False):
        """
        Load the workflow data and then save it to an h5ad file

        :param output_file_name: A path to the output file
        :type output_file_name: str
        :param to_sparse: Convert the data to sparse prior to saving it
        :type to_sparse: bool
        """

        self.read_expression()
        self.read_tfs()

        if to_sparse and not self.data.is_sparse:
            self.data.to_sparse()

        self.data.to_h5ad(self.output_path(output_file_name), compression="gzip")

    @staticmethod
    def _create_null_prior(gene_names, tf_names):
        """
        Create a prior data matrix that is all 0s
        :param gene_names: Anything that can be used as an index for a dataframe
        :param tf_names: list
        :return priors: pd.DataFrame
        """
        return pd.DataFrame(0, index=gene_names, columns=tf_names)

    @staticmethod
    def dataframe_split(data_frame, remove_columns):
        """
        Take a dataframe and extract specific columns. Return the dataframe, minus those columns, and a second
        dataframe which is only those columns.
        :param data_frame: pd.DataFrame
        :param remove_columns: list(str)
        :return data_frame, data_frame_two: pd.DataFrame, pd.DataFrame
        """

        data_frame_two = data_frame.loc[:, remove_columns].copy()
        data_frame = data_frame.drop(remove_columns, axis=1)
        return data_frame, data_frame_two

    @staticmethod
    def loaded_file_info(df_name, df):

        Debug.vprint(df_name + " loaded {sh}".format(sh=df.shape), level=2)
        Debug.vprint(df_name + " index: " + str(df.index[0]) + " ...", level=2)
        Debug.vprint(df_name + " columns: " + str(df.columns[0]) + " ...", level=2)


class WorkflowBase(WorkflowBaseLoader):
    """
    WorkflowBase handles crossvalidation, shuffling, and validating priors and gold standards
    """
    # Flags to control splitting priors into a prior/gold-standard set
    split_gold_standard_for_crossvalidation = False
    cv_split_ratio = None
    cv_split_axis = 0

    # Flags to control shuffle and noise baselines
    shuffle_prior_axis = None
    make_data_noise = None
    _data_is_noise = False

    # The random seed for sampling, etc
    random_seed = 42

    # The number of inference bootstraps to run
    num_bootstraps = 2

    # Use the Intel MKL libraries for matrix multiplication
    use_mkl = None

    # Multiprocessing controller
    initialize_mp = True
    multiprocessing_controller = None

    # Prior manager
    prior_manager = ManagePriors

    # Result processing & model metrics
    _result_processor_driver = ResultsProcessor
    gold_standard_filter_method = "keep_all_gold_standard"
    metric = "combined"

    # Output results in an InferelatorResults object
    results = None

    def __init__(self):
        super(WorkflowBase, self).__init__()
        # Get environment variables
        self.get_environmentals()

    def set_crossvalidation_parameters(self, split_gold_standard_for_crossvalidation=None, cv_split_ratio=None,
                                       cv_split_axis=None):
        """
        Set parameters for crossvalidation.

        :param split_gold_standard_for_crossvalidation: Boolean flag indicating if the gold standard should be
            split. Must be set to True for other crossvalidation settings to have an effect. Defaults to False.
        :type split_gold_standard_for_crossvalidation: bool
        :param cv_split_ratio: The proportion of the gold standard which should be retained for scoring. The rest will
            be used to train the model. Must be set betweeen 0 and 1.
        :type cv_split_ratio: float
        :param cv_split_axis: How to split the gold standard. If 0, split genes; this will take all the data for certain
            genes and keep it in the gold standard. These genes will be removed from the prior. If 1, split regulators;
            this will take all the data for certain regulators and keep it in the gold standard. These regulators will
            be removed from the prior. Splitting regulators is inadvisable. If None, the prior will be replaced with a
            downsampled gold standard. Setting this to 0 is generally the best choice.
            Defaults to None.
        :type cv_split_axis: int, None

        """

        self._set_without_warning("split_gold_standard_for_crossvalidation", split_gold_standard_for_crossvalidation)
        self._set_with_warning("cv_split_ratio", cv_split_ratio)
        self._set_with_warning("cv_split_axis", cv_split_axis)

        if not split_gold_standard_for_crossvalidation and (cv_split_axis is not None or cv_split_ratio is not None):
            warnings.warn("The split_gold_standard_for_crossvalidation flag is not set. Other options may be ignored")

    def set_shuffle_parameters(self, shuffle_prior_axis=None, make_data_noise=None):
        """
        Set parameters for shuffling labels on a prior axis. This is useful to establish a baseline.

        :param shuffle_prior_axis: The axis for shuffling prior labels. 0 shuffles gene labels. 1 shuffles regulator
            labels. None means labels will not be shuffled. Defaults to None.
        :type shuffle_prior_axis: int, None
        :param make_data_noise: Replace loaded data with simulated data that is entirely random. This retains type;
            integer data remains integer, float remains float. Gene distributions should be centered around the
            mean of gene expression in the original data, but is otherwise random.
        """
        self._set_with_warning("shuffle_prior_axis", shuffle_prior_axis)
        self._set_with_warning("make_data_noise", make_data_noise)

    def set_postprocessing_parameters(self, gold_standard_filter_method=None, metric=None):
        """
        Set parameters for the postprocessing engine

        :param gold_standard_filter_method: A flag that determines if the gold standard should be shrunk to the
            size of the produced model. "overlap" will only score on overlap between the gold standard and the
            inferred gene regulatory network. "keep_all_gold_standard" will score on the entire gold standard.
            Defaults to "keep_all_gold_standard".
        :type gold_standard_filter_method: str
        :param metric: The model metric to use for scoring. Supports "precision-recall", "mcc", "f1", and "combined"
            Defaults to "combined".
        :type metric: str
        """

        self._set_with_warning("gold_standard_filter_method", gold_standard_filter_method)
        self._set_with_warning("metric", metric)

    @staticmethod
    def set_output_file_names(network_file_name="", confidence_file_name="", nonzero_coefficient_file_name="",
                              pdf_curve_file_name="", curve_data_file_name=""):
        """
        Set output file names. File names that end in '.gz' will be gzipped.

        :param network_file_name: Long-format network TSV file with TF->Gene edge information.
            Default is "network.tsv".
        :type network_file_name: str
        :param confidence_file_name: Genes x TFs TSV with confidence scores for each edge.
            Default is "combined_confidences.tsv"
        :type confidence_file_name: str
        :param nonzero_coefficient_file_name: Genes x TFs TSV with the number of non-zero model coefficients for each
            edge. Default is None (this file is not produced).
        :type nonzero_coefficient_file_name: str
        :param pdf_curve_file_name: PDF file with plotted curve(s). Default is "combined_metrics.pdf".
        :type pdf_curve_file_name: str
        :param curve_data_file_name: TSV file with the data used to plot curves.
            Default is None (this file is not produced).
        :type curve_data_file_name: str
        """

        if network_file_name != "":
            InferelatorResults.network_file_name = network_file_name
        if confidence_file_name != "":
            InferelatorResults.confidence_file_name = confidence_file_name
        if nonzero_coefficient_file_name != "":
            InferelatorResults.threshold_file_name = nonzero_coefficient_file_name
        if pdf_curve_file_name != "":
            InferelatorResults.curve_file_name = pdf_curve_file_name
        if curve_data_file_name != "":
            InferelatorResults.curve_data_file_name = curve_data_file_name

    def set_run_parameters(self, num_bootstraps=None, random_seed=None, use_mkl=None):
        """
        Set parameters used during runtime

        :param num_bootstraps: The number of bootstraps to run. Defaults to 2.
        :type num_bootstraps: int
        :param random_seed: The random number seed to use. Defaults to 42.
        :type random_seed: int
        :param use_mkl: A flag to indicate if the intel MKL library should be used for matrix multiplication
        :type use_mkl: bool
        """

        self._set_without_warning("num_bootstraps", num_bootstraps)
        self._set_without_warning("random_seed", random_seed)
        self._set_without_warning("use_mkl", use_mkl)

    def initialize_multiprocessing(self):
        """
        Register the multiprocessing controller if set and run .connect()
        """
        if self.multiprocessing_controller is not None:
            MPControl.set_multiprocess_engine(self.multiprocessing_controller)
        MPControl.connect()

    def get_environmentals(self):
        """
        Load environmental variables into class variables
        """
        for k, v in slurm_envs(SBATCH_VARS_FOR_WORKFLOW).items():
            setattr(self, k, v)

    def startup(self):
        """
        Startup by preprocessing all data into a ready format for regression.
        """

        DotProduct.set_mkl(self.use_mkl)

        if self.initialize_mp and not MPControl.is_initialized:
            self.initialize_multiprocessing()

        self.startup_run()
        self.startup_finish()

    def startup_run(self):
        """
        Execute any data preprocessing necessary before regression. Startup_run is mostly for reading in data
        """
        raise NotImplementedError  # implement in subclass

    def startup_finish(self):
        """
        Execute any data preprocessing necessary before regression. Startup_finish is mostly for preprocessing data
        prior to regression
        """
        raise NotImplementedError  # implement in subclass

    def run(self):
        """
        Execute workflow, after all configuration.
        """
        raise NotImplementedError  # implement in subclass

    def process_priors_and_gold_standard(self):
        """
        Run split, shuffle, etc called for by the workflow flags
        This also filters the expression matrix to the list of genes to model
        """

        # Split gold standard for cross-validation
        if self.split_gold_standard_for_crossvalidation:
            self.priors_data, self.gold_standard = self.prior_manager.cross_validate_gold_standard(self.priors_data,
                                                                                                   self.gold_standard,
                                                                                                   self.cv_split_axis,
                                                                                                   self.cv_split_ratio,
                                                                                                   self.random_seed)

        # Filter priors to a list of regulators
        if self.tf_names is not None:
            self.priors_data = self.prior_manager.filter_to_tf_names_list(self.priors_data, self.tf_names)
        else:
            self.tf_names = self.priors_data.columns.tolist()

        # Filter priors and expression to a list of genes
        self.filter_to_gene_list()

        # Shuffle prior labels
        if self.shuffle_prior_axis is not None:
            self.priors_data = self.prior_manager.shuffle_priors(self.priors_data,
                                                                 self.shuffle_prior_axis,
                                                                 self.random_seed)

        # Check for duplicates or whatever
        self.priors_data, self.gold_standard = self.prior_manager.validate_priors_gold_standard(self.priors_data,
                                                                                                self.gold_standard)

    def filter_to_gene_list(self):
        """
        Filter the priors and expression matrix to just genes in gene_metadata
        """

        Debug.vprint("Trimming expression matrix", level=1)
        self.data.trim_genes(trim_gene_list=self.gene_names)
        self.priors_data = self.prior_manager.filter_priors_to_genes(self.priors_data, self.data.gene_names)

    def align_priors_and_expression(self):
        """
        Align prior to the expression matrix
        """
        self.priors_data = self.prior_manager.align_priors_to_expression(self.priors_data, self.data.gene_names)
        self.data_white_noise()

    def data_white_noise(self):
        """
        Replace data with white noise data
        """

        if self.make_data_noise and not self._data_is_noise:
            make_data_noisy(self.data, random_seed=self.random_seed)
            self._data_is_noise = True

    def get_bootstraps(self):
        """
        Generate sequence of bootstrap parameter objects for run.
        """
        col_range = range(self._num_obs)
        random_state = np.random.RandomState(seed=self.random_seed)
        return random_state.choice(col_range, size=(self.num_bootstraps, self._num_obs)).tolist()

    def emit_results(self, betas, rescaled_betas, gold_standard, priors):
        """
        Output result report(s) for workflow run.
        """
        raise NotImplementedError  # implement in subclass

    @staticmethod
    def is_master():
        """
        Return True if this is the master thread
        """
        return MPControl.is_master

    def create_output_dir(self):
        """
        Set a default output_dir if nothing is set. Create the path if it doesn't exist.
        """
        if self.output_dir is None:
            new_path = datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
            self.output_dir = InferelatorDataLoader.make_path_safe(os.path.join(self.input_dir, new_path))
        else:
            self.output_dir = InferelatorDataLoader.make_path_safe(self.output_dir)

        try:
            os.makedirs(os.path.expanduser(self.output_dir))
        except FileExistsError:
            pass

    def create_task(self, **kwargs):
        """
        Create a task data object
        """
        raise NotImplementedError("This workflow does not support multiple tasks")


def _factory_build_inferelator(regression=_RegressionWorkflowMixin, workflow=WorkflowBase):
    """
    This is the factory method to create workflow classes that combine preprocessing and postprocessing (from workflow)
    with a regression method (from regression)

    :param regression: RegressionWorkflow subclass
        A class object which implements the run_regression and run_bootstrap methods for a specific regression strategy
    :param workflow: WorkflowBase subclass
        A class object which implements the necessary data loading and preprocessing to create design & response data
        for the regression strategy, and then the postprocessing to turn regression betas into a network
    :return RegressWorkflow:
        This returns an uninstantiated class which is the multi-inheritance result of both the regression workflow and
        the preprocessing/postprocessing workflow
    """

    use_mtl_regression = False

    # Decide which preprocessing/postprocessing workflow to use
    # String arguments are parsed for convenience in the run script
    if is_string(workflow):
        workflow = workflow.lower()
        if workflow == "base":
            workflow_class = WorkflowBase
        elif workflow == "tfa":
            from inferelator.tfa_workflow import TFAWorkFlow
            workflow_class = TFAWorkFlow
        elif workflow == "amusr" or workflow == "multitask":
            from inferelator.amusr_workflow import MultitaskLearningWorkflow
            workflow_class = MultitaskLearningWorkflow
            use_mtl_regression = True
        elif workflow == "single-cell":
            from inferelator.single_cell_workflow import SingleCellWorkflow
            workflow_class = SingleCellWorkflow
        elif workflow == "velocity":
            from inferelator.velocity_workflow import VelocityWorkflow
            workflow_class = VelocityWorkflow
        else:
            raise ValueError("{val} is not a string that can be mapped to a workflow class".format(val=workflow))
    # Or just use a workflow class directly
    elif inspect.isclass(workflow) and issubclass(workflow, WorkflowBase):
        workflow_class = workflow
    else:
        raise ValueError("Workflow must be a string that maps to a workflow class or an actual workflow class")

    # Decide which regression workflow to use
    # Return just the workflow if regression is set to None
    if regression is None:
        return workflow_class
    # String arguments are parsed for convenience in the run script
    elif is_string(regression):
        regression = regression.lower()
        if regression == "base":
            regression_class = _RegressionWorkflowMixin
        elif regression == "bbsr" and not use_mtl_regression:
            from inferelator.regression.bbsr_python import BBSRRegressionWorkflowMixin
            regression_class = BBSRRegressionWorkflowMixin
        elif regression == "elasticnet" and not use_mtl_regression:
            from inferelator.regression.elasticnet_python import ElasticNetWorkflowMixin
            regression_class = ElasticNetWorkflowMixin
        elif regression == "amusr":
            from inferelator.regression.amusr_regression import AMUSRRegressionWorkflowMixin
            regression_class = AMUSRRegressionWorkflowMixin
        elif regression == "bbsr-by-task" or (regression == "bbsr" and use_mtl_regression):
            from inferelator.regression.bbsr_multitask import BBSRByTaskRegressionWorkflowMixin
            regression_class = BBSRByTaskRegressionWorkflowMixin
        elif regression == "elasticnet-by-task" or (regression == "elasticnet" and use_mtl_regression):
            from inferelator.regression.elasticnet_python import ElasticNetByTaskRegressionWorkflowMixin
            regression_class = ElasticNetByTaskRegressionWorkflowMixin
        elif regression == "stars-by-task" or (regression == "stars" and use_mtl_regression):
            from inferelator.regression.stability_selection import StARSWorkflowByTaskMixin
            regression_class = StARSWorkflowByTaskMixin
        elif regression == "stars":
            from inferelator.regression.stability_selection import StARSWorkflowMixin
            regression_class = StARSWorkflowMixin
        elif regression == "sklearn" and not use_mtl_regression:
            from inferelator.regression.sklearn_regression import SKLearnWorkflowMixin
            regression_class = SKLearnWorkflowMixin
        elif regression == "sklearn" and use_mtl_regression:
            from inferelator.regression.sklearn_regression import SKLearnByTaskMixin
            regression_class = SKLearnByTaskMixin
        else:
            raise ValueError("{val} is not a string that can be mapped to a regression class".format(val=regression))
    # Or just use a regression class directly
    elif inspect.isclass(regression) and issubclass(regression, _RegressionWorkflowMixin):
        regression_class = regression
    else:
        raise ValueError("Regression must be a string that maps to a regression class or an actual regression class")

    class RegressWorkflow(regression_class, workflow_class):
        regression_type = regression_class

    return RegressWorkflow


def inferelator_workflow(regression=_RegressionWorkflowMixin, workflow=WorkflowBase):
    """
    Create and instantiate an Inferelator workflow.

    :param regression: A class object which implements the run_regression and run_bootstrap methods for a specific
        regression strategy. This can be provided as a string.

        "base" loads a non-functional regression stub.

        "bbsr" loads Bayesian Best Subset Regression.

        "elasticnet" loads Elastic Net Regression.

        "sklearn" loads scikit-learn Regression.

        "stars" loads the StARS stability Regression.

        "amusr" loads AMuSR Regression. This requires multitask workflow.

        "bbsr-by-task" loads Bayesian Best Subset Regression for multiple tasks. This requires multitask workflow.

        "elasticnet-by-task" loads Elastic Net Regression for multiple tasks. This requires multitask workflow.

        Defaults to "base".
    :type regression: str, RegressionWorkflow subclass
    :param workflow: A class object which implements the necessary data loading and preprocessing to create design &
        response data for the regression strategy, and then the postprocessing to turn regression betas into a network.
        This can be provided as a string.

        "base" loads a non-functional workflow stub.

        "tfa" loads the TFA-based workflow.

        "single-cell" loads the Single Cell TFA-based workflow.

        "multitask" loads the multitask workflow.

        Defaults to "base".
    :type workflow: str, WorkflowBase subclass
    :return: This returns an initialized object which has both the regression workflow and the
        preprocessing/postprocessing workflow. This object can then have settings assigned to it, and can be run
        with `.run()`
    :rtype: Workflow instance
    """
    return _factory_build_inferelator(regression=regression, workflow=workflow)()



"""
Run Single Cell Network Inference. This is the same network inference with some extra preprocessing functionality.
"""

import numpy as np

from inferelator.utils import Validator as check
from inferelator import tfa_workflow
from inferelator.preprocessing import single_cell
from inferelator import utils

PREPROCESSING_FUNCTIONS = {"log2": single_cell.log2_data,
                           "ln": single_cell.ln_data,
                           "log10": single_cell.log10_data,
                           "ftt": single_cell.tf_sqrt_data}


class SingleCellWorkflow(tfa_workflow.TFAWorkFlow):
    """
    SingleCellWorkflow has some additional preprocessing prior to calculating TFA and running regression
    """
    # Single-cell expression data manipulations
    count_minimum = None  # float

    # Preprocessing workflow holder
    preprocessing_workflow = None

    # Do not use a design-response driver
    drd_driver = None

    def set_count_minimum(self, count_minimum=None):
        """
        Set the minimum count value for each gene (averaged over all samples)

        :param count_minimum: The mean expression value which is required to retain a gene for modeling. Data that
            has already been normalized should probably be filtered during normalization, not now.
            Defaults to None (disabled).
        :type count_minimum: float
        """

        self.count_minimum = count_minimum

    def add_preprocess_step(self, fun, **kwargs):
        """
        Add a preprocessing step after count filtering but before calculating TFA or regression.

        :param fun: Preprocessing function. Can be provided as a string or as a function in `preprocessing.single_cell`.

            "log10" will take the log10 of pseudocounts

            "ln" will take the natural log of pseudocounts

            "log2" will take the log2 of pseudocounts

            "fft" will do the Freeman-Tukey transform

        :type fun: str, `preprocessing.single_cell` function
        :param kwargs: Additional arguments to the preprocessing function
        """
        if self.preprocessing_workflow is None:
            self.preprocessing_workflow = []

        if utils.is_string(fun) and fun.lower() in PREPROCESSING_FUNCTIONS:
            self.preprocessing_workflow.append((PREPROCESSING_FUNCTIONS[fun], kwargs))
        elif utils.is_string(fun) and fun.lower() not in PREPROCESSING_FUNCTIONS:
            raise ValueError("Unable to translate {f} into a function to call".format(f=fun))
        else:
            self.preprocessing_workflow.append((fun, kwargs))

    def startup_finish(self):
        # Preprocess the single-cell data based on the preprocessing steps added to the workflow
        self.data_white_noise()
        self.single_cell_normalize()

        super(SingleCellWorkflow, self).startup_finish()

    def single_cell_normalize(self):
        """
        Single cell normalization. Requires expression_matrix to be all numeric, and to be [N x G].
        Executes all preprocessing workflow steps from the preprocessing_workflow list that's set by the
        add_preprocess_step() class function
        """

        single_cell.filter_genes_for_count(self.data, count_minimum=self.count_minimum)

        if self.preprocessing_workflow is not None:
            for sc_func, sc_kwargs in self.preprocessing_workflow:
                sc_kwargs['random_seed'] = self.random_seed
                sc_func(self.data, **sc_kwargs)

        num_nonfinite, name_nonfinite = self.data.non_finite
        if num_nonfinite > 0:
            utils.Debug.vprint("These genes have non-finite values: " + " ".join(name_nonfinite), level=0)
            raise ValueError("NaN values have been introduced into the expression matrix by normalization")



"""
Run Multitask Network Inference with TFA-AMuSR.
"""
import copy
import gc
import warnings

from inferelator.utils import Debug
from inferelator import workflow
from inferelator import single_cell_workflow
from inferelator.regression import amusr_regression
from inferelator.postprocessing.results_processor_mtl import ResultsProcessorMultiTask

TRANSFER_ATTRIBUTES = ['count_minimum', 'preprocessing_workflow', 'input_dir', 'make_data_noise']
NON_TASK_ATTRIBUTES = ["gold_standard_file", "random_seed", "num_bootstraps"]


class MultitaskLearningWorkflow(single_cell_workflow.SingleCellWorkflow):
    """
    Class that implements multitask learning. Handles loading and validation of multiple data packages.
    """

    _regulator_expression_filter = "intersection"
    _target_expression_filter = "union"

    # Task-specific data
    _n_tasks = None
    _task_design = None
    _task_response = None
    _task_bootstraps = None
    _task_priors = None
    _task_names = None
    _task_objects = None

    task_results = None

    # Axis labels to keep
    _targets = None
    _regulators = None

    # Multi-task result processor
    _result_processor_driver = ResultsProcessorMultiTask

    @property
    def _num_obs(self):
        if self._task_objects is not None:
            return sum([t if t is not None else 0 for t in map(lambda x: x._num_obs, self._task_objects)])
        else:
            return None

    @property
    def _num_genes(self):
        if self._task_objects is not None:
            return max([t if t is not None else 0 for t in map(lambda x: x._num_genes, self._task_objects)])
        else:
            return None

    @property
    def _num_tfs(self):
        if self._task_objects is not None:
            return max([t if t is not None else 0 for t in map(lambda x: x._num_tfs, self._task_objects)])
        else:
            return None

    def set_task_filters(self, regulator_expression_filter=None, target_expression_filter=None):
        """
        Set the filtering criteria for regulators and targets between tasks

        :param regulator_expression_filter:
            "union" includes regulators which are present in any task,
            "intersection" includes regulators which are present in all tasks
        :type regulator_expression_filter: str, optional
        :param target_expression_filter:
            "union" includes targets which are present in any task,
            "intersection" includes targets which are present in all tasks
        :type target_expression_filter: str, optional
        """

        self._set_without_warning("_regulator_expression_filter", regulator_expression_filter)
        self._set_without_warning("_target_expression_filter", target_expression_filter)

    def startup_run(self):
        """
        Load data.

        This is called when `.startup()` is run. It is not necessary to call separately.
        """

        self.get_data()
        self.validate_data()

    def get_data(self):
        # Task data has expression & metadata and may have task-specific files for anything else
        self._load_tasks()

        # Priors, gold standard, tf_names, and gene metadata will be loaded if set
        self.read_tfs()
        self.read_priors()
        self.read_genes()

    def startup_finish(self):
        """
        Process task data and priors.

        This is called when `.startup()` is run. It is not necessary to call separately.
        """

        # Make sure tasks are set correctly
        self._process_default_priors()
        self._process_task_priors()
        self._process_task_data()

    def create_task(self, task_name=None, input_dir=None, expression_matrix_file=None, meta_data_file=None,
                    tf_names_file=None, priors_file=None, gene_names_file=None, gene_metadata_file=None,
                    workflow_type="single-cell", **kwargs):
        """
        Create a task object and set any arguments to this function as attributes of that task object. TaskData objects
        are stored internally in _task_objects.

        :param task_name: A descriptive name for this task
        :type task_name: str
        :param input_dir: A path containing the input files
        :type input_dir: str
        :param expression_matrix_file: Path to the expression data
        :type expression_matrix_file: str
        :param meta_data_file: Path to the meta data
        :type meta_data_file: str, optional
        :param tf_names_file: Path to a list of regulator names to include in the model
        :type tf_names_file: str
        :param priors_file: Path to a prior data file
        :type priors_file: str
        :param gene_metadata_file: Path to a genes annotation file
        :type gene_metadata_file: str, optional
        :param gene_names_file: Path to a list of genes to include in the model (optional)
        :type gene_names_file: str, optional
        :param workflow_type: The type of workflow for data preprocessing.
            "tfa" uses the TFA workflow,
            "single-cell" uses the Single-Cell TFA workflow
        :type workflow_type: str, `inferelator.BaseWorkflow` subclass
        :param kwargs: Any additional arguments are assigned to the task object.
        :return: Returns a task reference which can be additionally modified by calling any valid Workflow function to
            set task parameters
        :rtype: TaskData instance
        """

        # Create a TaskData object from a workflow and set the formal arguments into it
        task_object = create_task_data_object(workflow_class=workflow_type)
        task_object.task_name = task_name
        task_object.input_dir = input_dir if input_dir is not None else self.input_dir
        task_object.expression_matrix_file = expression_matrix_file
        task_object.meta_data_file = meta_data_file
        task_object.tf_names_file = tf_names_file
        task_object.priors_file = priors_file
        task_object.gene_names_file = gene_names_file
        task_object.gene_metadata_file = gene_metadata_file

        # Warn if there is an attempt to set something that isn't supported
        msg = "Task-specific {} is not supported. This setting will be ignored. Set this in the parent workflow."
        for bad in NON_TASK_ATTRIBUTES:
            if bad in kwargs:
                del kwargs[bad]
                warnings.warn(msg.format(bad))

        # Pass forward any kwargs (raising errors if they're for attributes that don't exist)
        for attr, val in kwargs.items():
            if hasattr(task_object, attr):
                setattr(task_object, attr, val)
                task_object.str_attrs.append(attr)
            else:
                raise ValueError("Argument {attr} cannot be set as an attribute".format(attr=attr))

        if self._task_objects is None:
            self._task_objects = [task_object]
        else:
            self._task_objects.append(task_object)

        Debug.vprint(str(task_object), level=0)

        return task_object

    def _load_tasks(self):
        """
        Run load_task_data in all the TaskData objects created with create_task
        """
        if self._task_objects is None:
            raise ValueError("Tasks have not been created with .create_task()")

        for tobj in self._task_objects:
            # Transfer attributes from parent if they haven't been set in the task
            for attr in TRANSFER_ATTRIBUTES:
                try:
                    if getattr(self, attr) is not None and getattr(tobj, attr) is None:
                        setattr(tobj, attr, getattr(self, attr))
                except AttributeError:
                    pass

            # Set the random seed in the task to the same as the parent
            tobj.random_seed = self.random_seed

            # Set the num_bootstraps in the task to the same as the parent
            tobj.num_bootstraps = self.num_bootstraps

        # Run load_task_data and create a list of lists of TaskData objects
        # This allows a TaskData object to copy and split itself if needed
        self._task_objects = [tobj.get_data() for tobj in self._task_objects]

        # Flatten the list
        self._task_objects = [tobj for tobj_list in self._task_objects for tobj in tobj_list]
        self._n_tasks = len(self._task_objects)

    def validate_data(self):
        """
        Make sure that the data that's loaded is acceptable

        This is called when `.startup()` is run. It is not necessary to call separately.

        :raises ValueError: Raises a ValueError if any tasks have invalid priors or gold standard structures
        """
        if self.gold_standard is None:
            raise ValueError("A gold standard must be provided to `gold_standard_file` in MultiTaskLearningWorkflow")

        # Check to see if there are any tasks which don't have priors
        no_priors = sum(map(lambda x: x.priors_data is None, self._task_objects))
        if no_priors > 0 and self.priors_data is None:
            raise ValueError("{n} tasks have no priors (no default prior is set)".format(n=no_priors))

    def _process_default_priors(self):
        """
        Process the default priors in the parent workflow for crossvalidation or shuffling
        """

        priors = self.priors_data if self.priors_data is not None else self.gold_standard.copy()

        # Crossvalidation
        if self.split_gold_standard_for_crossvalidation:
            priors, self.gold_standard = self.prior_manager.cross_validate_gold_standard(priors, self.gold_standard,
                                                                                         self.cv_split_axis,
                                                                                         self.cv_split_ratio,
                                                                                         self.random_seed)
        # Filter to regulators
        if self.tf_names is not None:
            priors = self.prior_manager.filter_to_tf_names_list(priors, self.tf_names)

        # Filter to targets
        if self.gene_names is not None:
            priors = self.prior_manager.filter_priors_to_genes(priors, self.gene_names)

        # Shuffle labels
        if self.shuffle_prior_axis is not None:
            priors = self.prior_manager.shuffle_priors(priors, self.shuffle_prior_axis, self.random_seed)

        # Reset the priors_data in the parent workflow if it exists
        self.priors_data = priors if self.priors_data is not None else None

    def _process_task_priors(self):
        """
        Process & align the default priors for crossvalidation or shuffling
        """
        for task_obj in self._task_objects:

            # Set priors if task-specific priors are not present
            if task_obj.priors_data is None and self.priors_data is None:
                raise ValueError("No priors exist in the main workflow or in tasks")
            elif task_obj.priors_data is None:
                task_obj.priors_data = self.priors_data.copy()

            # Set gene names if task-specific gene names is not present
            if task_obj.gene_names is None:
                task_obj.gene_names = copy.copy(self.gene_names)

            # Set tf_names if task-specific tf names are not present
            if task_obj.tf_names is None:
                task_obj.tf_names = copy.copy(self.tf_names)

            # Process priors in the task data
            task_obj.process_priors_and_gold_standard(gold_standard=self.gold_standard,
                                                      cv_flag=self.split_gold_standard_for_crossvalidation,
                                                      cv_axis=self.cv_split_axis,
                                                      shuffle_priors=self.shuffle_prior_axis)

    def _process_task_data(self):
        """
        Preprocess the individual task data using the TaskData worker into task design and response data. Set
        self.task_design, self.task_response, self.task_bootstraps with lists which contain
        DataFrames.

        Also set self.regulators and self.targets with pd.Indexes that correspond to the genes and tfs to model
        This is chosen based on the filtering strategy set in self.target_expression_filter and
        self.regulator_expression_filter
        """
        self._task_design, self._task_response = [], []
        self._task_bootstraps, self._task_names, self._task_priors = [], [], []
        targets, regulators = [], []

        # Iterate through a list of TaskData objects holding data
        for task_id, task_obj in enumerate(self._task_objects):
            # Get task name from Task
            task_name = task_obj.task_name if task_obj.task_name is not None else str(task_id)

            task_str = "Processing task #{tid} [{t}] {sh}"
            Debug.vprint(task_str.format(tid=task_id, t=task_name, sh=task_obj.data.shape), level=1)

            # Run the preprocessing workflow
            task_obj.startup_finish()

            # Put the processed data into lists
            self._task_design.append(task_obj.design)
            self._task_response.append(task_obj.response)
            self._task_bootstraps.append(task_obj.get_bootstraps())
            self._task_names.append(task_name)
            self._task_priors.append(task_obj.priors_data)

            regulators.append(task_obj.design.gene_names)
            targets.append(task_obj.response.gene_names)

            task_str = "Processing task #{tid} [{t}] complete [{sh} & {sh2}]"
            Debug.vprint(task_str.format(tid=task_id, t=task_name, sh=task_obj.design.shape,
                                         sh2=task_obj.response.shape), level=1)

        self._targets = amusr_regression.filter_genes_on_tasks(targets, self._target_expression_filter)
        self._regulators = amusr_regression.filter_genes_on_tasks(regulators, self._regulator_expression_filter)

        Debug.vprint("Processed data into design/response [{g} x {k}]".format(g=len(self._targets),
                                                                              k=len(self._regulators)), level=0)

        # Clean up the TaskData objects and force a cyclic collection
        del self._task_objects
        gc.collect()

        # Make sure that the task data files have the correct columns
        for d in self._task_design:
            d.trim_genes(remove_constant_genes=False, trim_gene_list=self._regulators)

        for r in self._task_response:
            r.trim_genes(remove_constant_genes=False, trim_gene_list=self._targets)

    def emit_results(self, betas, rescaled_betas, gold_standard, priors_data):
        """
        Output result report(s) for workflow run.

        This is called when `.startup()` is run. It is not necessary to call separately.
        """
        if self.is_master():
            self.create_output_dir()
            rp = self._result_processor_driver(betas, rescaled_betas, filter_method=self.gold_standard_filter_method,
                                               metric=self.metric)
            rp.tasks_names = self._task_names
            self.results = rp.summarize_network(self.output_dir, gold_standard, self._task_priors)
            self.task_results = rp.tasks_networks
            return self.results
        else:
            return None


def create_task_data_object(workflow_class="single-cell"):
    return create_task_data_class(workflow_class=workflow_class)()


def create_task_data_class(workflow_class="single-cell"):
    task_parent = workflow._factory_build_inferelator(regression="base", workflow=workflow_class)

    class TaskData(task_parent):
        """
        TaskData is a workflow object which only loads and preprocesses data from files.
        """

        task_name = None
        tasks_from_metadata = False
        meta_data_task_column = None

        task_workflow_class = str(workflow_class)

        str_attrs = ["input_dir", "expression_matrix_file", "tf_names_file", "meta_data_file", "priors_file"]

        def __str__(self):
            """
            Create a printable report of the settings in this TaskData object

            :return: Settings in str_attrs in a printable string
            :rtype: str
            """

            task_str = "{n}:\n\tWorkflow Class: {cl}\n".format(n=self.task_name, cl=self.task_workflow_class)
            for attr in self.str_attrs:
                try:
                    task_str += "\t{attr}: {val}\n".format(attr=attr, val=getattr(self, attr))
                except AttributeError:
                    task_str += "\t{attr}: Nonexistant\n".format(attr=attr)
            return task_str

        def __init__(self):
            if self._file_format_settings is None:
                self._file_format_settings = dict()

        def initialize_multiprocessing(self):
            """
            Don't do anything with multiprocessing in this object
            """
            pass

        def startup(self):
            raise NotImplementedError

        def startup_run(self):
            raise NotImplementedError

        def get_data(self):
            """
            Load all the data and then return a list of references to TaskData objects
            There will be multiple objects returned if tasks_from_metadata is set.
            If tasks_from_metadata is not set, the list contains only this task (self)

            :return: List of TaskData objects with loaded data
            :rtype: list(TaskData)
            """
            Debug.vprint("Loading data for task {task_name}".format(task_name=self.task_name))
            super(TaskData, self).get_data()

            if self.tasks_from_metadata:
                return self.separate_tasks_by_metadata()
            else:
                return [self]

        def validate_data(self):
            """
            Don't validate data in TaskData. The parent workflow will check.
            """

            pass

        def set_run_parameters(self):
            """
            Set parameters used during runtime
            """

            warnings.warn("Task-specific `num_bootstraps` and `random_seed` is not supported. Set on parent workflow.")

        def process_priors_and_gold_standard(self, gold_standard=None, cv_flag=None, cv_axis=None, shuffle_priors=None):
            """
            Make sure that the priors for this task are correct
            """

            gold_standard = self.gold_standard if gold_standard is None else gold_standard
            cv_flag = self.split_gold_standard_for_crossvalidation if cv_flag is None else cv_flag
            cv_axis = self.cv_split_axis if cv_axis is None else cv_axis
            shuffle_priors = self.shuffle_prior_axis if shuffle_priors is None else shuffle_priors

            # Remove circularity from the gold standard
            if cv_flag:
                self.priors_data, _ = self.prior_manager._remove_prior_circularity(self.priors_data, gold_standard,
                                                                                   split_axis=cv_axis)

            if self.tf_names is not None:
                self.priors_data = self.prior_manager.filter_to_tf_names_list(self.priors_data, self.tf_names)

            # Filter priors and expression to a list of genes
            self.filter_to_gene_list()

            # Shuffle prior labels
            if shuffle_priors is not None:
                self.priors_data = self.prior_manager.shuffle_priors(self.priors_data, shuffle_priors, self.random_seed)

            if min(self.priors_data.shape) == 0:
                raise ValueError("Priors for task {n} have an axis of length 0".format(n=self.task_name))

        def separate_tasks_by_metadata(self, meta_data_column=None):
            """
            Take a single expression matrix and break it into multiple dataframes based on meta_data. Return a list of
            TaskData objects which have the task-specific data loaded into them

            :param meta_data_column: Meta_data column which corresponds to task ID
            :type meta_data_column: str
            :return new_task_objects: List of the TaskData objects with only one task's data each
            :rtype: list(TaskData)

            """

            if self.data is None:
                raise ValueError("No data has been loaded prior to `separate_tasks_by_metadata`")

            meta_data_column = meta_data_column if meta_data_column is not None else self.meta_data_task_column
            if meta_data_column is None:
                raise ValueError("tasks_from_metadata is set but meta_data_task_column is not")
            elif meta_data_column not in self.data.meta_data:
                msg = "meta_data_task_column is not found in task {t}".format(t=str(self))
                raise ValueError(msg)

            new_task_objects = list()
            tasks = self.data.meta_data[meta_data_column].unique().tolist()
            Debug.vprint("Creating {n} tasks from metadata column {col}".format(n=len(tasks), col=meta_data_column),
                         level=0)

            # Remove data references from self
            data = self.data
            self.data = None

            for task in tasks:
                # Copy this object
                task_obj = copy.deepcopy(self)

                # Get an index of the stuff to keep
                task_idx = data.meta_data[meta_data_column] == task

                # Reset expression matrix, metadata, and task_name in the copy
                task_obj.data = data.subset_copy(row_index=task_idx)
                task_obj.data.name = task
                task_obj.task_name = task
                new_task_objects.append(task_obj)

            Debug.vprint("Separated data into {ntask} tasks".format(ntask=len(new_task_objects)), level=0)

            return new_task_objects

    return TaskData



from inferelator import default
import os

# Maintain python 2 compatibility
try:
    from itertools import izip as zip
except ImportError:
    pass

from inferelator.distributed import AbstractController

from dask import distributed

try:
    DEFAULT_LOCAL_DIR = os.environ['TMPDIR']
except KeyError:
    DEFAULT_LOCAL_DIR = 'dask-worker-space'


class DaskK8Controller(AbstractController):
    """
    The DaskK8Controller class launches a local dask cluster and connects as a client

    The map functionality is deliberately not implemented; dask-specific multiprocessing functions are used instead
    """

    _controller_name = "dask-k8"
    _controller_dask = True

    is_master = True
    client = None

    ## Dask controller variables ##

    # The dask cluster object
    local_cluster = None

    # Settings for dask workers
    processes = 2
    local_dir = DEFAULT_LOCAL_DIR

    @classmethod
    def connect(cls, *args, **kwargs):
        """
        Setup local cluster
        """

        # kwargs["n_workers"] = kwargs.pop("n_workers", cls.processes)
        # kwargs["threads_per_worker"] = kwargs.pop("threads_per_worker", 1)
        kwargs["processes"] = kwargs.pop("processes", True)

        # Ugly hack because dask-jobqueue changed this keyword arg
        local_directory = kwargs.pop("local_dir", None)
        local_directory = kwargs.pop("local_directory", None) if local_directory is None else local_directory
        kwargs["local_directory"] = local_directory if local_directory is not None else cls.local_dir

        cls.local_cluster = distributed.LocalCluster()
        cls.client = distributed.Client()
        return True

    @classmethod
    def shutdown(cls):
        cls.client.close()
        # cls.local_cluster.close()

    @classmethod
    def map(cls, func, *args, **kwargs):
        raise NotImplementedError

    @classmethod
    def set_processes(cls, process_count):
        """
        Set the number of dask workers to use
        :param process_count: int
        :return:
        """
        cls.processes = process_count

    @classmethod
    def sync_processes(self, *args, **kwargs):
        """
        This is a thing for KVS. Just return True.
        """
        return True

    @classmethod
    def check_cluster_state(cls, *args, **kwargs):
        """
        This is a thing for dask. Just return True.
        """
        return True



from inferelator import amusr_workflow
from inferelator import workflow
from inferelator.regression.base_regression import _RegressionWorkflowMixin
from inferelator.postprocessing.results_processor import ResultsProcessor
from inferelator.tests.artifacts.test_data import TestDataSingleCellLike, TEST_DATA, TEST_DATA_SPARSE
from inferelator.utils import InferelatorData

import pandas as pd
import numpy as np


class NoOutputRP(ResultsProcessor):

    def summarize_network(self, output_dir, gold_standard, priors):
        return super(NoOutputRP, self).summarize_network(None, gold_standard, priors)


# Factory method to spit out a puppet workflow
def create_puppet_workflow(regression_class=_RegressionWorkflowMixin,
                           base_class=workflow.WorkflowBase,
                           result_processor_class=NoOutputRP):

    puppet_parent = workflow._factory_build_inferelator(regression=regression_class, workflow=base_class)

    class PuppetClass(puppet_parent):
        """
        Standard workflow except it takes all the data as references to __init__ instead of as filenames on disk or
        as environment variables, and returns the model AUPR and edge counts without writing files (unless told to)
        """

        write_network = True
        network_file_name = None
        pr_curve_file_name = None
        initialize_mp = False

        def __init__(self, data, prior_data, gs_data):
            self.data = data
            self.priors_data = prior_data
            self.gold_standard = gs_data
            super(PuppetClass, self).__init__()

        def startup_run(self):
            # Skip all of the data loading
            self.process_priors_and_gold_standard()

        def create_output_dir(self, *args, **kwargs):
            pass

    return PuppetClass


class TaskDataStub(amusr_workflow.create_task_data_class(workflow_class="single-cell")):
    priors_data = TestDataSingleCellLike.priors_data
    tf_names = TestDataSingleCellLike.tf_names

    meta_data_task_column = "Condition"
    tasks_from_metadata = True

    task_name = "TestStub"
    task_workflow_type = "single-cell"

    def __init__(self, sparse=False):
        self.data = TEST_DATA.copy() if not sparse else TEST_DATA_SPARSE.copy()
        super(TaskDataStub, self).__init__()

    def get_data(self):
        if self.tasks_from_metadata:
            return self.separate_tasks_by_metadata()
        else:
            return [self]


class FakeDRD:
    def __init__(self, *args, **kwargs):
        pass

    def run(self, expr, meta):
        return expr, expr, expr

    def validate_run(self, meta):
        return True


class FakeWriter(object):
    def writerow(self, *args, **kwargs):
        pass


class FakeRegressionMixin(_RegressionWorkflowMixin):

    def run_regression(self):
        beta = [pd.DataFrame(np.array([[0, 1], [0.5, 0.05]]), index=['gene1', 'gene2'], columns=['tf1', 'tf2'])]
        beta_resc = [pd.DataFrame(np.array([[0, 1], [1, 0.05]]), index=['gene1', 'gene2'], columns=['tf1', 'tf2'])]
        return beta, beta_resc

    def run_bootstrap(self, bootstrap):
        return True


class FakeResultProcessor:
    network_data = None

    def __init__(self, *args, **kwargs):
        pass

    def summarize_network(self, *args, **kwargs):
        return 1, 0, 0



import os
import pandas as pd
import numpy as np
import gzip
from inferelator import utils
from inferelator.utils import Validator as check
from inferelator.postprocessing import (GOLD_STANDARD_COLUMN, CONFIDENCE_COLUMN, TARGET_COLUMN, REGULATOR_COLUMN,
                                        TN, FN, TP, FP)


class RankSummingMetric(object):
    """
    This class takes a data set that has some rankable values and a gold standard for which elements of that data set
    are true and calculates confidence
    """

    # Metric name
    name = "Confidences"

    # Filter methods to align gold standard and confidences
    filter_method = None
    filter_method_lookup = {'overlap': 'filter_to_overlap', 'keep_all_gold_standard': 'filter_to_left_size'}

    # Data (wide)
    rankable_data = None
    gold_standard = None
    all_confidences = None

    # Processed data (long)
    confidence_data = None
    filtered_data = None

    # File name
    curve_file_name = None

    def __init__(self, rankable_data, gold_standard, filter_method='keep_all_gold_standard'):
        """
        Take rankable data and process it into confidence scores which are stored in this object
        :param rankable_data: list(pd.DataFrame) [B x [G x K]]
            A list of numeric dataframes (with identical axes)
        :param gold_standard: pd.DataFrame [G x K]
            A dataframe which corresponds to known, gold-standard data
        :param filter_method: str
            The method of aligning the

        """

        # Get the filtering method
        assert check.argument_enum(filter_method, self.filter_method_lookup.keys())
        self.filter_method = getattr(self, self.filter_method_lookup[filter_method])

        # Explicitly cast the gold standard data to a boolean array [0,1]
        gold_standard = (gold_standard != 0).astype(int)
        self.gold_standard = gold_standard

        # Calculate confidences based on the ranked data
        self.all_confidences = self.compute_combined_confidences(rankable_data)

        # Convert the confidence data to long format
        confidence_data = utils.melt_and_reindex_dataframe(self.all_confidences, CONFIDENCE_COLUMN,
                                                           idx_name=TARGET_COLUMN, col_name=REGULATOR_COLUMN)

        # Attach the gold standard
        confidence_data = self.attach_gs_to_confidences(confidence_data, gold_standard)

        # Sort by confidence (descending) and reset the index
        self.confidence_data = confidence_data.sort_values(by=CONFIDENCE_COLUMN, ascending=False, na_position='last')
        self.confidence_data.reset_index(inplace=True)

        # Filter the gold standard and confidences down to a format that can be directly compared
        utils.Debug.vprint("GS: {gs} edges, Confidences: {conf} edges".format(gs=gold_standard.shape[0],
                                                                              conf=self.confidence_data.shape[0]),
                           level=0)

        self.filtered_data = self.filter_method(GOLD_STANDARD_COLUMN, CONFIDENCE_COLUMN, self.confidence_data).copy()
        utils.Debug.vprint("Filtered data to {e} edges".format(e=self.filtered_data.shape[0], level=0))

    def score(self):
        raise NotImplementedError

    def all_scores(self):
        return dict([tuple(self.score())])

    @classmethod
    def all_names(cls):
        return tuple([cls.name])

    def auc(self):
        raise NotImplementedError

    def confidence_dataframe(self):
        return self.confidence_data

    def curve_dataframe(self):
        raise NotImplementedError

    def output_curve_pdf(self, output_dir, file_name=None, dpi=300, figsize=(6, 4)):

        file_name = self.curve_file_name if file_name is None else file_name

        # Plot the curve
        ax = self.output_curve(figsize=figsize)
        fig = ax.get_figure()

        # If there's a file name set, make the output file
        if file_name is not None and output_dir is not None:
            # Save the plot and close
            self.save_figure(os.path.join(output_dir, file_name), fig, dpi=dpi)

        return ax

    def output_curve(self, ax=None, figsize=(6, 4)):
        raise NotImplementedError

    @staticmethod
    def save_figure(file_name, fig, dpi=300):
        """
        Save figure. Gzip if the file name ends with .gz

        :param file_name: File name
        :type file_name: str
        :param fig: Figure object
        :type fig: matplotlib.Figure
        :param dpi: DPI of output
        :type dpi: int
        """
        if file_name is None:
            return
        elif file_name.lower().endswith(".gz"):
            with gzip.open(file_name, mode='w') as fh:
                fig.savefig(fh, dpi=dpi)
        else:
            fig.savefig(file_name, dpi=dpi)

    @staticmethod
    def attach_gs_to_confidences(confidence_data, gold_standard):
        """
        Outer join the gold standard into the confidence data

        :param confidence_data: pd.DataFrame [G*K x n]
        :param gold_standard: pd.DataFrame [G x K]
        :return:
        """

        gold_standard = utils.melt_and_reindex_dataframe(gold_standard, GOLD_STANDARD_COLUMN, idx_name=TARGET_COLUMN,
                                                         col_name=REGULATOR_COLUMN)

        return confidence_data.join(gold_standard, how='outer', on=[TARGET_COLUMN, REGULATOR_COLUMN])

    @staticmethod
    def compute_combined_confidences(rankable_data):
        """
        Calculate confidences based on ranking value in all of the data frames and summing the ranks
        :param rankable_data: list(pd.DataFrame [M x N])
        :return combine_conf: pd.DataFrame [M x N]
        """
        # Create an 0s dataframe shaped to the data to be ranked
        combine_conf = pd.DataFrame(np.zeros(rankable_data[0].shape),
                                    index=rankable_data[0].index,
                                    columns=rankable_data[0].columns)

        for replicate in rankable_data:
            # Flatten and rank based on the beta error reductions
            ranked_replicate = np.reshape(pd.DataFrame(replicate.values.flatten()).rank().values, replicate.shape)
            # Sum the rankings for each bootstrap
            combine_conf += ranked_replicate

        # Convert rankings to confidence values
        min_element = min(combine_conf.values.flatten())
        combine_conf = (combine_conf - min_element) / (len(rankable_data) * combine_conf.size - min_element)
        return combine_conf

    @staticmethod
    def compute_confusion_matrix(data, rank_col=CONFIDENCE_COLUMN, gs_col=GOLD_STANDARD_COLUMN):

        # Copy off just needed columns
        data = data[[rank_col, gs_col]].copy()

        # Force sort if necessary
        if not data.loc[~pd.isnull(data[rank_col]), rank_col].is_monotonic_decreasing:
            _reindex = data.index.copy()
            data.sort_values(by=rank_col, ascending=False, na_position='last', inplace=True)
        else:
            _reindex = None

        # Get indices for testable edges
        valid_gs_idx = ~pd.isnull(data[gs_col])

        # Find the edges that are in the gold standard
        df = data.loc[valid_gs_idx, [rank_col, gs_col]].copy()
        df[gs_col] = df[gs_col].astype(bool)

        # Calculate cumulative confusion matrix at each row
        df[TP] = (df[gs_col]).astype(int).cumsum()
        df[FP] = (~df[gs_col]).astype(int).cumsum()
        df[TN] = pd.Series((~df[gs_col].iloc[::-1]).astype(int).cumsum(),
                           index=df.index).shift(-1, fill_value=0)
        df[FN] = pd.Series((df[gs_col].iloc[::-1]).astype(int).cumsum(),
                           index=df.index).shift(-1, fill_value=0)

        # Handle ties
        RankSummingMetric.transform_column(df, rank_col, TP, 'max')
        RankSummingMetric.transform_column(df, rank_col, FP, 'max')
        RankSummingMetric.transform_column(df, rank_col, TN, 'min')
        RankSummingMetric.transform_column(df, rank_col, FN, 'min')

        # Stick confusion results back onto the data and return it
        data[TP], data[FP], data[TN], data[FN] = pd.NA, pd.NA, pd.NA, pd.NA
        data.loc[valid_gs_idx, [TP, FP, TN, FN]] = df[[TP, FP, TN, FN]]

        if _reindex is not None:
            data = data.reindex(_reindex)

        return data

    @staticmethod
    def filter_to_left_size(left_column, right_column, data):
        # Return data where one column (left) is not NA
        return data.dropna(subset=[left_column])

    @staticmethod
    def filter_to_overlap(left_column, right_column, data):
        # Return data where both columns are not NA
        return data.dropna(subset=[left_column, right_column])

    @staticmethod
    def transform_column(df, group_col, xform_col, xform):
        # Transform column based on a grouping column
        df[xform_col] = df[[group_col, xform_col]].groupby(group_col)[xform_col].transform(xform)



"""
KVSController is a wrapper for KVSClient that adds some useful functionality related to interprocess
communication.
It also keeps track of a bunch of SLURM related stuff that was previously workflow's problem.
"""

from kvsstcp import KVSClient

from inferelator.distributed import AbstractController
from inferelator.utils import Validator as check
from inferelator import utils
from inferelator import default

import os
import warnings
import collections
import tempfile

# Use cPickle in python 2
try:
    import cPickle as pickle
except ImportError:
    import pickle

# Shadow built-in zip with itertools.izip if this is python2 (This puts out a memory dumpster fire)
try:
    from itertools import izip as zip
except ImportError:
    pass

# KVS environment variables
SBATCH_VARS_FOR_KVS = ["SLURM_PROCID", "SLURM_NTASKS_PER_NODE", "SLURM_NTASKS", "SLURM_NODEID", "SLURM_JOB_NUM_NODES"]

# Which process is the boss
DEFAULT_MASTER = 0

# KVS Keys to use
COUNT = "kvs_count"
TMP_FILE_SYNC = "tmp_file_read"
POST_SYNC = "post_sync"
PILEUP_DATA = "data_pileup"
FINAL_DATA = "final_data"


class KVSController(AbstractController):

    # An active KVSClient object
    client = None
    chunk = 25
    _controller_name = "kvs"

    # Set from SLURM environment variables
    rank = None  # int
    tasks = None  # int
    node = None  # int
    cores = None  # int
    num_nodes = None  # int
    is_master = False  # bool

    @classmethod
    def connect(cls, *args, **kwargs):
        """
        Create a new KVS object with some object variables set to reflect the slurm environment
        """

        # Get local environment variables
        cls._get_env(master_rank=kwargs.pop("master_rank", DEFAULT_MASTER))

        # Connect to the host server by calling to KVSClient.__init__
        cls.client = KVSClient(*args, **kwargs)

    @classmethod
    def shutdown(cls):
        return cls.client.close()

    @classmethod
    def _get_env(cls, master_rank=DEFAULT_MASTER):
        """
        Get the SLURM environment variables that are set by sbatch at runtime.
        """
        for k, v in utils.slurm_envs(default.SBATCH_VARS_FOR_KVS).items():
            setattr(cls, k, v)
        if cls.rank == master_rank:
            cls.is_master = True
        else:
            cls.is_master = False

    @classmethod
    def own_check(cls, chunk=1, kvs_key='count'):
        if cls.is_master:
            return ownCheck(cls.client, 0, chunk=chunk, kvs_key=kvs_key)
        else:
            return ownCheck(cls.client, 1, chunk=chunk, kvs_key=kvs_key)

    @classmethod
    def master_remove_key(cls, kvs_key='count'):
        if cls.is_master:
            cls.client.get(kvs_key)

    @classmethod
    def set_processes(cls, process_count):
        """
        Set the number of dask workers to use
        :param process_count: int
        :return:
        """
        utils.Debug.vprint("The KVS engine does not support changing process numbers at runtime", level=0)

    @classmethod
    def sync_processes(cls, pref="", value=True):
        """
        Block all processes until they reach this point, then release them
        It may be wise to use unique prefixes if this is gonna get called rapidly so there's no collision
        Or not. I'm a comment, not a cop.
        :param pref: str
            Prefix attached to the KVS keys
        :param value: Anything you can pickle
            A value that will be checked for consistency between processes (if you set a different value in a
            process, a warning will be issued. This is mostly to check state if needed
        :return None:
        """

        wkey = pref + '_wait'
        ckey = pref + '_continue'

        # Every process puts a wait key up when it gets here
        cls.put_key(wkey, value)

        # The master pulls down the wait keys until it has all of them
        # Then it puts up a go key for each process

        if cls.is_master:
            for _ in range(cls.tasks):
                c_value = cls.get_key(wkey)
                if c_value != value:
                    warnings.warn("Sync warning: master {val_m} is not equal to client {val_c}".format(val_m=value,
                                                                                                       val_c=c_value))
            for _ in range(cls.tasks):
                cls.put_key(ckey, True)

        # Every process waits here until go keys are available
        cls.get_key(ckey)

    @classmethod
    def get_key(cls, key):
        """
        Wrapper for KVSClient get
        """
        return cls.client.get(key)

    @classmethod
    def put_key(cls, key, value):
        """
        Wrapper for KVSClient put
        """
        return cls.client.put(key, value)

    @classmethod
    def view_key(cls, key):
        """
        Wrapper for KVSClient view
        """
        return cls.client.view(key)

    @classmethod
    def map(cls, func, *args, **kwargs):
        """
        Map a function across iterable(s) and return a list of results

        :param func: function
            Mappable function
        :param args: iterable
            Iterator(s)
        :param tmp_file_path: path
            If this is not None, instead of putting data onto the KVS, data will be pickled to temp files and the
            path to the temp file will be put onto the KVS
        :param tell_children: bool
            If this is True, all processes will end up with the final data after assembly. If false, only the master
            will have the final data; others will return None
        :return results: list
        """

        tmp_file_path = kwargs.pop("tmp_file_path", None)
        tell_children = kwargs.pop("tell_children", True)

        assert check.argument_callable(func)
        assert check.argument_list_type(args, collections.Iterable)

        # Set up the multiprocessing
        owncheck = cls.own_check(chunk=cls.chunk, kvs_key=COUNT)
        results = dict()
        for pos, arg in enumerate(zip(*args)):
            if next(owncheck):
                results[pos] = func(*arg)

        # Process results and synchronize exit from the get call
        results = cls.process_results(results, tmp_file_path=tmp_file_path, tell_children=tell_children)
        cls.sync_processes(pref=POST_SYNC)
        cls.master_remove_key(kvs_key=COUNT)
        cls.master_remove_key(kvs_key=FINAL_DATA)
        return results

    @classmethod
    def process_results(cls, results, tmp_file_path=None, tell_children=True):
        """
        Pile up results from a get call
        :param results: dict
            A dict of results (keyed by position in a final list)
        :param tmp_file_path: path
            If this is not None, instead of putting data onto the KVS, data will be pickled to temp files and the
            path to the temp file will be put onto the KVS
        :param tell_children: bool
            If this is True, all processes will end up with the final data after assembly. If false, only the master
            will have the final data; others will return None
        :return: list
            A list of function results
        """

        if tmp_file_path is None:
            # Put the data in KVS
            cls.put_key(PILEUP_DATA, results)
        else:
            # Put the data in a pickled file
            temp_fd, temp_name = tempfile.mkstemp(prefix="kvs", dir=tmp_file_path)
            with os.fdopen(temp_fd, "wb") as temp:
                pickle.dump(results, temp, -1)
            cls.put_key(PILEUP_DATA, temp_name)

        if cls.is_master:
            # If this is the master thread, get all the data and pile it up
            pileup_results = dict()
            for _ in range(cls.tasks):
                if tmp_file_path is None:
                    pileup_results.update(cls.get_key(PILEUP_DATA))
                else:
                    temp_name = cls.get_key(PILEUP_DATA)
                    with open(temp_name, mode="rb") as temp:
                        pileup_results.update(pickle.load(temp))
                    os.remove(temp_name)

            # Put everything into a list based on the dict key
            pileup_list = [None] * len(pileup_results)
            for idx, val in pileup_results.items():
                pileup_list[idx] = val

            if tell_children and tmp_file_path is None:
                # Put the piled-up data into KVS
                cls.put_key(FINAL_DATA, pileup_list)
            elif tell_children:
                # Pute the piled-up data into a pickled file
                temp_fd, temp_name = tempfile.mkstemp(prefix="kvs", dir=tmp_file_path)
                with os.fdopen(temp_fd, "wb") as temp:
                    pickle.dump(pileup_list, temp, -1)
                cls.put_key(FINAL_DATA, temp_name)
            else:
                # Put a None onto KVS - only the master needs this data
                cls.put_key(FINAL_DATA, None)

        else:
            # If this is not the master thread, get the finalized data when the master is finished
            if tell_children and tmp_file_path is None:
                pileup_list = cls.view_key(FINAL_DATA)
            elif tell_children:
                with open(cls.view_key(FINAL_DATA), "rb") as temp:
                    pileup_list = pickle.load(temp)
            else:
                pileup_list = cls.view_key(FINAL_DATA)

        # Return the piled up data or wait until everyone is done so that the temp file can be deleted
        if tmp_file_path is None:
            return pileup_list
        else:
            cls.sync_processes(pref=TMP_FILE_SYNC)
            if cls.is_master:
                os.remove(cls.view_key(FINAL_DATA))
            return pileup_list


def ownCheck(kvs, rank, chunk=1, kvs_key='count'):
    """
    Generator
    :param kvs: KVSClient
        KVS object for server access
    :param chunk: int
        The size of the chunk given to each subprocess
    :param kvs_key: str
        The KVS key to increment (default is 'count')
    :yield: bool
        True if this process has dibs on whatever. False if some other process has claimed it first.
    """
    if rank == 0:
        kvs.put(kvs_key, 0)

    # Start at the baseline
    checks, lower, upper = 0, -1, -1

    while True:

        # Checks increments every loop
        # If it's greater than the upper bound, get a new lower bound from the KVS count
        # Set the new upper bound by adding chunk to lower
        # And then put the new upper bound back into KVS key

        if checks >= upper:
            lower = kvs.get(kvs_key)
            upper = lower + chunk
            kvs.put(kvs_key, upper)

        # Yield TRUE if this row belongs to this process and FALSE if it doesn't
        yield lower <= checks < upper
        checks += 1



import pandas as pd
import numpy as np

from inferelator import utils

"""
This file is all preprocessing functions. All functions must take positional arguments data (InferelatorData).
All other arguments must be keyword. All functions must change the data object in place and return nothing.

Normalization functions take batch_factor_column [str] as a kwarg
Imputation functions take random_seed [int] and output_file [str] as a kwarg 
"""


def normalize_expression_to_one(data, **kwargs):
    """

    :param data: InferelatorData [N x G]
    """

    utils.Debug.vprint('Normalizing UMI counts per cell ... ')

    # Divide each cell's raw count data by the total number of UMI counts for that cell
    data.divide(data.sample_counts, axis=1)


def normalize_expression_to_median(data, **kwargs):
    """

    :param data: InferelatorData [N x G]
    """

    target_value = np.median(data.sample_counts)
    data.divide((data.sample_counts / target_value), axis=1)


def normalize_medians_for_batch(data, batch_factor_column=None, **kwargs):
    """
    Calculate the median UMI count per cell for each batch. Transform all batches by dividing by a size correction
    factor, so that all batches have the same median UMI count (which is the median batch median UMI count)

    :param data: InferelatorData [N x G]
    :param batch_factor_column: str
        Which meta data column should be used to determine batches
    :return expression_matrix, meta_data: pd.DataFrame, pd.DataFrame
    """

    if batch_factor_column is None or batch_factor_column not in data.meta_data:
        _msg = "batch_factor_column must be set to one of the meta data columns"
        utils.Debug.vprint(_msg + ": {c}".format(c=data.meta_data.columns), level=0)
        raise ValueError(_msg)

    utils.Debug.vprint('Normalizing median counts between batches ... ')

    # Get UMI counts for each cell
    umi = data.sample_counts

    # Create a new dataframe with the UMI counts and the factor to batch correct on
    umi = pd.DataFrame({'umi': umi, batch_factor_column: data.meta_data[batch_factor_column]})

    # Group and take the median UMI count for each batch
    median_umi = umi.groupby(batch_factor_column).agg('median')

    # Convert to a correction factor based on the median of the medians
    median_umi = median_umi / median_umi['umi'].median()
    umi = umi.join(median_umi, on=batch_factor_column, how="left", rsuffix="_mod")

    # Apply the correction factor to all the data
    data.divide(umi['umi_mod'].values, axis=1)


def normalize_sizes_within_batch(data, batch_factor_column=None, **kwargs):
    """
    Calculate the median UMI count within each batch and then resize each sample so that each sample has the same total
    UMI count

    :param data: InferelatorData [N x G]
    :param batch_factor_column: str
        Which meta data column should be used to determine batches
    :return expression_matrix, meta_data: pd.DataFrame, pd.DataFrame
    """

    if batch_factor_column is None or batch_factor_column not in data.meta_data:
        _msg = "batch_factor_column must be set to one of the meta data columns"
        utils.Debug.vprint(_msg + ": {c}".format(c=data.meta_data.columns), level=0)
        raise ValueError(_msg)

    utils.Debug.vprint('Normalizing to median counts within batches ... ')

    # Get UMI counts for each cell
    umi = data.sample_counts

    # Create a new dataframe with the UMI counts and the factor to batch correct on
    umi = pd.DataFrame({'umi': umi, batch_factor_column: data.meta_data[batch_factor_column]})

    # Group and take the median UMI count for each batch
    median_umi = umi.groupby(batch_factor_column).agg('median')

    # Convert to a correction factor based on the median of the medians
    umi = umi.join(median_umi, on="Condition", how="left", rsuffix="_mod")
    umi['umi_mod'] = umi['umi'] / umi['umi_mod']

    # Apply the correction factor to all the data
    data.divide(umi['umi_mod'].values, axis=1)


def log10_data(data, **kwargs):
    """
    Transform the expression data by adding one and then taking log10. Ignore any kwargs.

    :param data: InferelatorData [N x G]
    """
    utils.Debug.vprint('Logging data [log10+1] ... ')
    data.transform(np.log10, add_pseudocount=True)


def log2_data(data, **kwargs):
    """
    Transform the expression data by adding one and then taking log2. Ignore any kwargs.

    :param data: InferelatorData [N x G]
    """
    utils.Debug.vprint('Logging data [log2+1]... ')
    data.transform(np.log2, add_pseudocount=True)


def ln_data(data, **kwargs):
    """
    Transform the expression data by adding one and then taking ln. Ignore any kwargs.

    :param data: InferelatorData [N x G]
    """
    utils.Debug.vprint('Logging data [ln+1]... ')
    data.transform(np.log1p, add_pseudocount=False)


def tf_sqrt_data(data, **kwargs):
    """
    Transform the expression data by sqrt(x) + sqrt(x+1) and restore sparsity with x - 1
    Based on Freeman & Tukey: https://projecteuclid.org/euclid.aoms/1177729756

    :param data: InferelatorData [N x G]
    """
    utils.Debug.vprint('Freeman-Tukey square root transformation [sqrt(x) + sqrt(x+1) - 1]... ')
    data.transform(lambda x: np.sqrt(x) + np.sqrt(x + 1) - 1)


def filter_genes_for_count(data, count_minimum=None):
    """
    Filter out any genes which have a variance of 0 by calling filter_genes_for_var. Filter out any genes which don't
    reach the minimum count (if count is not none)

    :param data: InferelatorData [N x G]
    :param count_minimum: num
        The minimum value per sample required to include any genes
   """

    if count_minimum is None:
        data.trim_genes(remove_constant_genes=True)
    else:
        count_minimum = count_minimum * data.shape[0]
        if np.min(data.expression_data.min(axis=0)) < 0:
            raise ValueError("Cannot use a count minimum on data with negative values")
        counts_per_gene = data.gene_counts
        if np.any(~np.isfinite(counts_per_gene)):
            raise ValueError("Non-finite values in count matrix")
        keep_genes = counts_per_gene >= count_minimum
        utils.Debug.vprint("Filtering {gn} genes [Count]".format(gn=data.shape[1] - np.sum(keep_genes)), level=1)
        data.trim_genes(remove_constant_genes=True, trim_gene_list=data.gene_names[keep_genes])
